import json
import dpath
from ytvs.constants import CONTENT_TYPE_JSON_HEADER, MOBILE_XHR_HEADERS
from ytvs.extractor import YoutubeBasicInfoExtractor
from .base_resource import Resource
from ytvs.utils import try_get, _search_attribute
from ..parameter import DataParameterBuilder


def _parse_transcript_data(raw_data):
    """
    YouTube 동영상 자막(raw_data)을 파싱.
    :param raw_data: 파싱할 자막 데이터.
    :return: 파싱된 자막 데이터.
    """
    transcrtipt_segment_list_renderer = _search_attribute(raw_data, 'transcriptSegmentListRenderer', 20)
    transcripts = YoutubeBasicInfoExtractor.extract_transcript_segment_list_renderer(transcrtipt_segment_list_renderer)
    transcript_count = len(transcripts)

    return {
        'kind': 'youtube#TranscriptResponse',
        'etag': None,
        'items': transcripts,
        'nextPageToken': None,
        'pageInfo': {
            'resultsPerPage': transcript_count,
            'totalResults': transcript_count
        },
        'prevPageToken': None
    }


class TranscriptResource(Resource):
    def get_transcript_by_video_id(self, video_id):
        """
        주어진 비디오 ID에 대한 자막을 조회.
        :param video_id: 자막을 요청할 비디오의 ID.
        :return: 파싱된 자막 데이터.
        """
        # 1. 영상 정보로부터 Transcript를 획득할 수 있는 Token을 획득
        transcript_token = self._request_video_transcript_token(video_id)

        # 2. transcript raw 데이터 획득
        transcript_contents_raw = self._request_video_transcript_contents(transcript_token)

        # 3. raw 데이터에서 필요한 데이터만 파싱
        return _parse_transcript_data(transcript_contents_raw)

    def _request_video_transcript_token(self, video_id: str):
        """
        주어진 비디오 ID에 대한 자막 토큰을 요청.
        :param video_id: 자막 토큰을 요청할 비디오의 ID.
        :return: 자막 토큰.
        """
        # 1. 연결 URL 생성
        endpoint = self.build_video_resource_url(video_id)
        video_html_text = self._client.fetch_url(endpoint).decode('utf-8')

        # 2. 데이터 파싱
        yt_initial_data = YoutubeBasicInfoExtractor.extract_yt_initial_data(video_html_text)
        transcript_token = dpath.get(
            yt_initial_data['init_data'],
            "**/content/continuationItemRenderer/continuationEndpoint/getTranscriptEndpoint/params",
            default=None,
        )
        return transcript_token

    def _request_video_transcript_contents(self, transcript_token):
        """
        주어진 토큰을 사용하여 YouTube 동영상의 자막 내용을 요청.
        :param transcript_token: 자막을 요청할 토큰.
        :return: 자막 내용의 Raw 데이터.
        """
        request_url = self.build_transcript_resource_url()
        data = json.dumps(DataParameterBuilder().set_param(transcript_token).build())
        content_bytes = self._client.fetch_url(
            request_url, headers=MOBILE_XHR_HEADERS + CONTENT_TYPE_JSON_HEADER, data=data)
        content_text = content_bytes.decode('utf-8')
        content_dict = json.loads(content_text)
        return content_dict



from dataclasses import dataclass, field
from typing import Optional, List

from .base import BaseModel
from .common import BaseApiResponse
from .mixins import DatetimeTimeMixin


@dataclass
class TranscriptSnippet(BaseModel, DatetimeTimeMixin):
    """
    A class representing the transcript snippet info.

    Refer: unknown
    """
    target_id: Optional[str] = field(default=None, repr=False)
    text: Optional[str] = field(default=None, repr=False)
    start_ms: Optional[str] = field(default=None, repr=False)
    end_ms: Optional[str] = field(default=None, repr=False)
    tracking_params: Optional[str] = field(default=None, repr=False)


@dataclass
class TranscriptResponse(BaseApiResponse):
    """
    A class representing the transcript's retrieve response info.

    Refer: unknown
    """

    items: Optional[List[TranscriptSnippet]] = field(default=None, repr=False)



"""
    There are channel banner related models

    References: https://developers.google.com/youtube/v3/docs/channelBanners#properties
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel


@dataclass
class ChannelBanner(BaseModel):
    """
    A class representing the channel banner's info.

    References: https://developers.google.com/youtube/v3/docs/channelBanners#resource
    """

    kind: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None, repr=False)
    url: Optional[str] = field(default=None)



"""
    Activities resource implementation
"""
from typing import Optional, Union

from pyyoutube.error import PyYouTubeException, ErrorMessage, ErrorCode
from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import ActivityListResponse
from pyyoutube.utils.params_checker import enf_parts


class ActivitiesResource(Resource):
    """An activity resource contains information about an action that a particular channel,
    or user, has taken on YouTube.

    References: https://developers.google.com/youtube/v3/docs/activities
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        channel_id: Optional[str] = None,
        mine: Optional[bool] = None,
        max_results: Optional[int] = None,
        page_token: Optional[str] = None,
        published_after: Optional[str] = None,
        published_before: Optional[str] = None,
        region_code: Optional[str] = None,
        return_json: bool = False,
        **kwargs,
    ) -> Union[dict, ActivityListResponse]:
        """Returns a list of channel activity events that match the request criteria.

        Args:
            parts:
                Comma-separated list of one or more activity resource properties.
            channel_id:
                The channelId parameter specifies a unique YouTube channel ID.
            mine:
                This parameter can only be used in a properly authorized request. Set this parameter's value
                to true to retrieve a feed of the authenticated user's activities.
            max_results:
                The parameter specifies the maximum number of items that should be returned
                the result set.
                Acceptable values are 0 to 50, inclusive. The default value is 5.
            page_token:
                The parameter identifies a specific page in the result set that should be returned.
            published_after:
                The parameter specifies the earliest date and time that an activity could
                have occurred for that activity to be included in the API response.
            published_before:
                The parameter specifies the date and time before which an activity must
                have occurred for that activity to be included in the API response.
            region_code:
                The parameter instructs the API to return results for the specified country.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Activities data

        """
        params = {
            "part": enf_parts(resource="activities", value=parts),
            "maxResults": max_results,
            "pageToken": page_token,
            "publishedAfter": published_after,
            "publishedBefore": published_before,
            "regionCode": region_code,
            **kwargs,
        }
        if channel_id is not None:
            params["channelId"] = channel_id
        elif mine is not None:
            params["mine"] = mine
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of for_username, id, managedByMe or mine",
                )
            )

        response = self._client.request(path="activities", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else ActivityListResponse.from_dict(data)



"""
    Watermarks resource implementation.
"""
from typing import Optional, Union

from pyyoutube.resources.base_resource import Resource
from pyyoutube.media import Media, MediaUpload
from pyyoutube.models import Watermark


class WatermarksResource(Resource):
    def set(
        self,
        channel_id: str,
        body: Union[dict, Watermark],
        media: Media,
        on_behalf_of_content_owner: Optional[str] = None,
        **kwargs: Optional[dict],
    ) -> MediaUpload:
        """

        Args:
            channel_id:
                Specifies the YouTube channel ID for which the watermark is being provided.
            body:
                Provide watermark data in the request body. You can give dataclass or just a dict with data.
            media:
                Media for watermark image.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.
        Returns:
            Watermark set status.
        """
        params = {
            "channel_id": channel_id,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }

        # Build a media upload instance.
        media_upload = MediaUpload(
            client=self._client,
            resource="watermarks/set",
            media=media,
            params=params,
            body=body.to_dict_ignore_none(),
        )
        return media_upload

    def unset(
        self,
        channel_id: str,
        on_behalf_of_content_owner: Optional[str] = None,
        **kwargs: Optional[dict],
    ) -> bool:
        """Deletes a channel's watermark image.

        Args:
            channel_id:
                Specifies the YouTube channel ID for which the watermark is being unset.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many difference YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            watermark unset status.
        """
        params = {
            "channelId": channel_id,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }
        response = self._client.request(
            method="POST",
            path="watermarks/unset",
            params=params,
        )
        if response.ok:
            return True
        self._client.parse_response(response=response)



import logging
import re
# Unfavoured alias
import sys
import socket
import traceback
from datetime import datetime, timedelta
from typing import List, Any

from dateutil.relativedelta import relativedelta

from ytvs.compatible import COMPATIBLE_RE_PATTERN

logger = logging.getLogger(__name__)


compiled_regex_type = COMPATIBLE_RE_PATTERN

NO_DEFAULT = object()


def is_number_string(s):
    return s.isdigit()


def select_first_item_in_list(items: List[Any]):
    if len(items) > 0:
        item = items[0]
    else:
        item = None
    return item


def _search_attributes(obj, attr, max_depth=5):
    """
    Recursively find an attribute in a list or dictionary up to a specified depth.
    If one instance of the attribute is found, returns it directly.
    If multiple instances are found, returns them in a list.

    :param obj: The list or dictionary to search.
    :param attr: The attribute name to find.
    :param max_depth: The maximum depth to search.
    :return: The attribute value(s) if found, otherwise None.
    """

    def _find(obj, depth):
        if depth < 0:
            return []

        results = []

        if isinstance(obj, dict) and attr in obj:
            results.append(obj[attr])

        if isinstance(obj, (list, dict)):
            for element in obj.values() if isinstance(obj, dict) else obj:
                results.extend(_find(element, depth - 1))

        return results

    found = _find(obj, max_depth)
    if found:
        return found
    else:
        return None
    # if len(found) == 1:
    #     return found[0]
    # elif found:
    #     return found
    # else:
    #     return None


#
def _search_attribute(obj, attr, max_depth=1):
    """
    Recursively find an attribute in a list or dictionary up to a specified depth.

    :param obj: The list or dictionary to search.
    :param attr: The attribute name to find.
    :param max_depth: The maximum depth to search.
    :return: The attribute value if found, otherwise None.
    """
    if max_depth < 0:
        return None

    # Check if the current object itself has the attribute
    if isinstance(obj, dict) and attr in obj:
        return obj[attr]

    # If the object is a list or dictionary, iterate through its elements
    if isinstance(obj, (list, dict)):
        for element in obj.values() if isinstance(obj, dict) else obj:
            result = _search_attribute(element, attr, max_depth - 1)
            if result is not None:
                return result

    return None


def _search_regex(pattern, string, name, default=NO_DEFAULT, fatal=True, flags=0, group=None):
    """
    Perform a regex search on the given string, using a single or a list of
    patterns returning the first matching group.
    In case of failure return a default value or raise a WARNING or a
    RegexNotFoundError, depending on fatal, specifying the field name.
    """
    if isinstance(pattern, (str, str, compiled_regex_type)):
        mobj = re.search(pattern, string, flags)
    else:
        for p in pattern:
            mobj = re.search(p, string, flags)
            if mobj:
                break
    _name = name

    if mobj:
        if group is None:
            # return the first matching group
            return next(g for g in mobj.groups() if g is not None)
        elif isinstance(group, (list, tuple)):
            return tuple(mobj.group(g) for g in group)
        else:
            return mobj.group(group)
    elif default is not NO_DEFAULT:
        return default
    elif fatal:
        raise RegexNotFoundError('Unable to extract %s' % _name)
    else:
        return None


def is_korean_number(text):
    """
    Converts a Korean number with units like 천 (thousand) and 만 (ten thousand) into an integer.
    """
    # Regex to find number and unit
    match = re.search(r'(\d+(\.\d+)?)(천|만)', text)

    if match:
        return True
    else:
        return False


def convert_korean_number(text):
    """
    Converts a Korean number with units like 천 (thousand) and 만 (ten thousand) into an integer.
    """
    # Regex to find number and unit
    match = re.search(r'(\d+(\.\d+)?)(천|만)', text)

    if match:
        number = float(match.group(1))
        unit = match.group(3)

        # Convert based on Korean number system
        if unit == '만':
            number *= 10000
        elif unit == '천':
            number *= 1000

        return int(number)
    else:
        return None


def remove_non_numeric_korean(text):
    # 숫자, '백', '천', '만', '억'을 제외한 한글 삭제
    # 숫자와 콤마, 그리고 '백', '천', '만', '억'을 포함하는 정규표현식
    pattern = r'[^\d,백천만억]'

    # 정규표현식에 매칭되는 모든 한글을 삭제
    result = re.sub(pattern, '', text)

    return result


def extract_like_count(text):
    refined_text = text.replace(",", "")
    refined_text = remove_non_numeric_korean(refined_text)
    if not is_number_string(refined_text):
        logger.warning(f"Failed to extract like count from {text}")
        return 0
    return int(refined_text)


def extract_view_count(text):
    refined_text = text.replace(",", "")
    refined_text = remove_non_numeric_korean(refined_text)
    if not is_number_string(refined_text):
        logger.warning(f"Failed to extract view count from {text}")
        return 0
    return int(refined_text)
    # 숫자(콤마 포함)와 선택적으로 '백', '천', '만', '억'을 찾는 정규표현식
    # pattern = r'\d+(백|천|만|억)?'
    # pattern = r'\d'
    #
    # # 정규표현식에 매칭되는 모든 부분을 찾아 리스트로 반환
    # matches = re.findall(pattern, refined_text)
    #
    # # 찾은 부분들에서 불필요한 그룹을 제거하고 문자열로 변환
    # result = ' '.join(match[0] + match[2] for match in matches)
    # return result


def int_or_none(v, scale=1, default=None, get_attr=None, invscale=1):
    if get_attr:
        if v is not None:
            v = getattr(v, get_attr, None)
    if v in (None, ''):
        return default
    try:
        return int(v) * invscale // scale
    except (ValueError, TypeError, OverflowError):
        return default


def str_to_int(int_str):
    if isinstance(int_str, int):
        return int_str
    elif isinstance(int_str, str):
        int_str = re.sub(r'[,\.\+]', '', int_str)
        return int_or_none(int_str)


def try_get(src, getter, expected_type=None):
    if not isinstance(getter, (list, tuple)):
        getter = [getter]
    for get in getter:
        try:
            v = get(src)
        except (AttributeError, KeyError, TypeError, IndexError):
            pass
        else:
            if expected_type is None or isinstance(v, expected_type):
                return v


def number_str_abbr_to_int(int_str):
    try:
        count = convert_korean_number(int_str)
        if count is None:
            count = ''.join(filter(str.isdigit, int_str))
        return count
    except Exception as e:
        return None


class YoutubeDLError(Exception):
    """Base exception for YoutubeDL errors."""
    pass


class UnavailableVideoError(YoutubeDLError):
    """Unavailable Format exception.

    This exception will be thrown when a video is requested
    in a format that is not available for that video.
    """
    pass


class ExtractorError(YoutubeDLError):
    """Error during info extraction."""

    def __init__(self, msg, tb=None, expected=False, cause=None, video_id=None):
        """ tb, if given, is the original traceback (so that it can be printed out).
        If expected is set, this is a normal error message and most likely not a bug in youtube-dl.
        """

        if sys.exc_info()[0] in (compat_urllib_error.URLError, socket.timeout, UnavailableVideoError):
            expected = True
        if video_id is not None:
            msg = video_id + ': ' + msg
        if cause:
            msg += ' (caused by %r)' % cause
        if not expected:
            msg += ""
        super(ExtractorError, self).__init__(msg)

        self.traceback = tb
        self.exc_info = sys.exc_info()  # preserve original exception
        self.cause = cause
        self.video_id = video_id

    def format_traceback(self):
        if self.traceback is None:
            return None
        return ''.join(traceback.format_tb(self.traceback))


class RegexNotFoundError(ExtractorError):
    """Error when a regex didn't match"""
    pass



def remove_prefix_with_colon_in_date_text(text:str):
    # 정규표현식 패턴: 줄의 시작 부분부터 ':' 문자가 나오기 전까지의 모든 문자와 ':' 자체
    pattern = r'^[^:]+:'

    # 각 줄을 순회하면서 패턴에 해당하는 부분 제거
    lines = text.split('\n')
    result = []
    for line in lines:
        # 패턴이 있는지 확인
        if re.search(pattern, line):
            # 패턴을 제거하고 결과에 추가
            new_line = re.sub(pattern, '', line)
            result.append(new_line)
        else:
            # 패턴이 없는 경우 원래의 줄을 결과에 추가
            result.append(line)

    return '\n'.join(result)



def convert_relative_to_absolute_date(date):
    """
    날짜 포맷을 년.월.일. 형식으로 변경하는 함수
    네이버 뉴스의 경우
    0분전, 0시간 전, 0일전 등 7일 내 뉴스는
    년.월.일이 아닌 다른 포맷으로 표시되므로 날짜를 통일해주는 함수가 필요함
    """
    current_time = datetime.now()
    date = date.replace(" ", "")
    date = date.replace("(수정됨)", "")
    date = date.replace("스트리밍시간:", "")
    date = date.replace("스트리밍시작:", "")
    date = date.replace("최초공개:", "")
    date = remove_prefix_with_colon_in_date_text(date)
    if date.endswith('분전'):
        minutes = int(date[:-2])
        date = current_time - timedelta(minutes=minutes)

    elif date.endswith('시간전'):
        hours = int(date[:-3])
        date = current_time - timedelta(hours=hours)

    elif date.endswith('일전'):
        days = int(date[:-2])
        date = current_time - timedelta(days=days)

    elif date.endswith('주전'):
        weeks = int(date[:-2])
        date = current_time - timedelta(weeks=weeks)

    elif date.endswith('개월전'):
        months = int(date[:-3])
        date = current_time - relativedelta(months=months)

    elif date.endswith('년전'):
        years = int(date[:-2])
        date = current_time - relativedelta(years=years)
    else:
        try:
            date = datetime.strptime(date, '%Y.%m.%d.')
        except Exception as e:
            return None
    return date.strftime("%Y-%m-%d")



from .activity import *  # noqa
from .auth import AccessToken, UserProfile
from .caption import *  # noqa
from .category import *  # noqa
from .channel import *  # noqa
from .channel_banner import *  # noqa
from .channel_section import *  # noqa
from .comment import *  # noqa
from .comment_thread import *  # noqa
from .i18n import *  # noqa
from .member import *  # noqa
from .memberships_level import *  # noqa
from .playlist_item import *  # noqa
from .playlist import *  # noqa
from .search_result import *  # noqa
from .subscription import *  # noqa
from .video_abuse_report_reason import *  # noqa
from .video import *  # noqa
from .watermark import *  # noqa



"""
    These are video abuse report reason related models.
"""

from dataclasses import dataclass, field
from typing import Optional, List

from .base import BaseModel
from .common import BaseResource, BaseApiResponse


@dataclass
class SecondaryReason(BaseModel):
    """
    A class representing the video abuse report reason info

    Refer: https://developers.google.com/youtube/v3/docs/videoAbuseReportReasons#snippet.secondaryReasons
    """

    id: Optional[str] = field(default=None)
    label: Optional[str] = field(default=None, repr=True)


@dataclass
class VideoAbuseReportReasonSnippet(BaseModel):
    """
    A class representing the video abuse report snippet info

    Refer: https://developers.google.com/youtube/v3/docs/videoAbuseReportReasons#snippet
    """

    label: Optional[str] = field(default=None)
    secondaryReasons: Optional[List[SecondaryReason]] = field(default=None, repr=True)


@dataclass
class VideoAbuseReportReason(BaseResource):
    """
    A class representing the video abuse report info

    Refer: https://developers.google.com/youtube/v3/docs/videoAbuseReportReasons
    """

    snippet: Optional[VideoAbuseReportReasonSnippet] = field(default=None)


@dataclass
class VideoAbuseReportReasonListResponse(BaseApiResponse):
    """
    A class representing the I18n language list response info.

    Refer: https://developers.google.com/youtube/v3/docs/videoAbuseReportReasons/list#response_1
    """

    items: Optional[List[VideoAbuseReportReason]] = field(default=None, repr=False)



# YouTube API와 통신을 위한 기본 설정 값들을 정의

# 기본 API 키: YouTube API를 사용하기 위한 기본 키입니다.
DEFAULT_API_KEY = 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8'

# 모바일 사용자 에이전트: YouTube 요청을 모바일 기기에서 발생한 것처럼 표시하기 위한 사용자 에이전트입니다.
MOBILE_USER_AGENT = 'Mozilla/5.0 (Linux; Android 7.0; Redmi Note 4 Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Mobile Safari/537.36'
# 모바일 사용자 에이전트 헤더: HTTP 요청에 사용될 모바일 사용자 에이전트 헤더입니다.
MOBILE_USER_AGENT_HEADER = (('User-Agent', MOBILE_USER_AGENT),)

# 데스크톱 사용자 에이전트: YouTube 요청을 데스크톱 기기에서 발생한 것처럼 표시하기 위한 사용자 에이전트입니다.
DESKTOP_USER_AGENT = 'Mozilla/5.0 (Windows NT 6.1; rv:52.0) Gecko/20100101 Firefox/52.0'
# 데스크톱 사용자 에이전트 헤더: HTTP 요청에 사용될 데스크톱 사용자 에이전트 헤더입니다.
DESKTOP_USER_AGENT_HEADER = (('User-Agent', DESKTOP_USER_AGENT),)

# JSON 콘텐츠 타입 헤더: HTTP 요청 시 JSON 형식의 콘텐츠를 사용함을 나타냅니다.
CONTENT_TYPE_JSON_HEADER = (('Content-Type', 'application/json'),)

# 데스크톱 XHR 헤더: 데스크톱 환경에서 YouTube API와 통신할 때 사용되는 HTTP 헤더입니다.
DESKTOP_XHR_HEADERS = (
    ('Accept', '*/*'),
    ('Accept-Language', 'en-US,en;q=0.5'),
    ('X-YouTube-Client-Name', '1'),
    ('X-YouTube-Client-Version', '2.20180830'),
) + DESKTOP_USER_AGENT_HEADER

# 모바일 XHR 헤더: 모바일 환경에서 YouTube API와 통신할 때 사용되는 HTTP 헤더입니다.
MOBILE_XHR_HEADERS = (
    ('Accept', '*/*'),
    ('Accept-Language', 'en-US,en;q=0.5'),
    ('X-YouTube-Client-Name', '2'),
    ('X-YouTube-Client-Version', '2.20180830'),
) + MOBILE_USER_AGENT_HEADER


"""
    Channel banners resource implementation.
"""
from typing import Optional

from pyyoutube.resources.base_resource import Resource
from pyyoutube.media import Media, MediaUpload


class ChannelBannersResource(Resource):
    """A channelBanner resource contains the URL that you would use to set a newly uploaded image as
    the banner image for a channel.

    References: https://developers.google.com/youtube/v3/docs/channelBanners
    """

    def insert(
        self,
        media: Media,
        on_behalf_of_content_owner: Optional[str] = None,
        **kwargs: Optional[dict],
    ) -> MediaUpload:
        """Uploads a channel banner image to YouTube.

        Args:
            media:
                Banner media data.
            on_behalf_of_content_owner:
                The onBehalfOfContentOwner parameter indicates that the request's authorization
                credentials identify a YouTube CMS user who is acting on behalf of the content
                owner specified in the parameter value. This parameter is intended for YouTube
                content partners that own and manage many different YouTube channels. It allows
                content owners to authenticate once and get access to all their video and channel
                data, without having to provide authentication credentials for each individual channel.
                The CMS account that the user authenticates with must be linked to the specified YouTube content owner.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Channel banner data.
        """
        params = {"onBehalfOfContentOwner": on_behalf_of_content_owner, **kwargs}
        # Build a media upload instance.
        media_upload = MediaUpload(
            client=self._client,
            resource="channelBanners/insert",
            media=media,
            params=params,
        )
        return media_upload



"""
    Membership levels resource implementation.
"""
from typing import Optional, Union

from pyyoutube.models import MembershipsLevelListResponse
from pyyoutube.resources.base_resource import Resource
from pyyoutube.utils.params_checker import enf_parts


class MembershipLevelsResource(Resource):
    """A membershipsLevel resource identifies a pricing level managed by the creator that authorized the API request.

    References: https://developers.google.com/youtube/v3/docs/membershipsLevels
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, MembershipsLevelListResponse]:
        """Lists membership levels for the channel that authorized the request.

        Args:
            parts:
                Comma-separated list of one or more channel resource properties.
                Accepted values: id,snippet
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Membership levels data.

        """
        params = {
            "part": enf_parts(resource="membershipsLevels", value=parts),
            **kwargs,
        }
        response = self._client.request(path="membershipsLevels", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else MembershipsLevelListResponse.from_dict(data)



"""
    These are video related models.
"""
from dataclasses import dataclass, field
from typing import Optional, List

import isodate
from isodate import ISO8601Error

from pyyoutube.error import ErrorCode, ErrorMessage, PyYouTubeException
from .base import BaseModel
from .common import (
    BaseApiResponse,
    BaseTopicDetails,
    BaseResource,
    Localized,
    Player,
    Thumbnails,
)
from .mixins import DatetimeTimeMixin


@dataclass
class RegionRestriction(BaseModel):
    """
    A class representing the video content details region restriction info

    Refer: https://developers.google.com/youtube/v3/docs/videos#contentDetails.regionRestriction
    """

    allowed: List[str] = field(default=None)
    blocked: List[str] = field(default=None, repr=False)


# TODO get detail rating description
class ContentRating(BaseModel):
    """
    A class representing the video content rating info.

    Refer: https://developers.google.com/youtube/v3/docs/videos#contentDetails.contentRating
    """

    acbRating: Optional[str] = field(default=None, repr=False)
    agcomRating: Optional[str] = field(default=None, repr=False)
    anatelRating: Optional[str] = field(default=None, repr=False)
    bbfcRating: Optional[str] = field(default=None, repr=False)
    bfvcRating: Optional[str] = field(default=None, repr=False)
    bmukkRating: Optional[str] = field(default=None, repr=False)
    catvRating: Optional[str] = field(default=None, repr=False)
    catvfrRating: Optional[str] = field(default=None, repr=False)
    cbfcRating: Optional[str] = field(default=None, repr=False)
    cccRating: Optional[str] = field(default=None, repr=False)
    cceRating: Optional[str] = field(default=None, repr=False)
    chfilmRating: Optional[str] = field(default=None, repr=False)
    chvrsRating: Optional[str] = field(default=None, repr=False)
    cicfRating: Optional[str] = field(default=None, repr=False)
    cnaRating: Optional[str] = field(default=None, repr=False)
    cncRating: Optional[str] = field(default=None, repr=False)
    csaRating: Optional[str] = field(default=None, repr=False)
    cscfRating: Optional[str] = field(default=None, repr=False)
    czfilmRating: Optional[str] = field(default=None, repr=False)
    djctqRating: Optional[str] = field(default=None, repr=False)
    djctqRatingReasons: List[str] = field(default=None, repr=False)
    ecbmctRating: Optional[str] = field(default=None, repr=False)
    eefilmRating: Optional[str] = field(default=None, repr=False)
    egfilmRating: Optional[str] = field(default=None, repr=False)
    eirinRating: Optional[str] = field(default=None, repr=False)
    fcbmRating: Optional[str] = field(default=None, repr=False)
    fcoRating: Optional[str] = field(default=None, repr=False)
    fpbRating: Optional[str] = field(default=None, repr=False)
    fpbRatingReasons: List[str] = field(default=None, repr=False)
    fskRating: Optional[str] = field(default=None, repr=False)
    grfilmRating: Optional[str] = field(default=None, repr=False)
    icaaRating: Optional[str] = field(default=None, repr=False)
    ifcoRating: Optional[str] = field(default=None, repr=False)
    ilfilmRating: Optional[str] = field(default=None, repr=False)
    incaaRating: Optional[str] = field(default=None, repr=False)
    kfcbRating: Optional[str] = field(default=None, repr=False)
    kijkwijzerRating: Optional[str] = field(default=None, repr=False)
    kmrbRating: Optional[str] = field(default=None, repr=False)
    lsfRating: Optional[str] = field(default=None, repr=False)
    mccaaRating: Optional[str] = field(default=None, repr=False)
    mccypRating: Optional[str] = field(default=None, repr=False)
    mcstRating: Optional[str] = field(default=None, repr=False)
    mdaRating: Optional[str] = field(default=None, repr=False)
    medietilsynetRating: Optional[str] = field(default=None, repr=False)
    mekuRating: Optional[str] = field(default=None, repr=False)
    mibacRating: Optional[str] = field(default=None, repr=False)
    mocRating: Optional[str] = field(default=None, repr=False)
    moctwRating: Optional[str] = field(default=None, repr=False)
    mpaaRating: Optional[str] = field(default=None, repr=False)
    mpaatRating: Optional[str] = field(default=None, repr=False)
    mtrcbRating: Optional[str] = field(default=None, repr=False)
    nbcRating: Optional[str] = field(default=None, repr=False)
    nfrcRating: Optional[str] = field(default=None, repr=False)
    nfvcbRating: Optional[str] = field(default=None, repr=False)
    nkclvRating: Optional[str] = field(default=None, repr=False)
    oflcRating: Optional[str] = field(default=None, repr=False)
    pefilmRating: Optional[str] = field(default=None, repr=False)
    resorteviolenciaRating: Optional[str] = field(default=None, repr=False)
    rtcRating: Optional[str] = field(default=None, repr=False)
    rteRating: Optional[str] = field(default=None, repr=False)
    russiaRating: Optional[str] = field(default=None, repr=False)
    skfilmRating: Optional[str] = field(default=None, repr=False)
    smaisRating: Optional[str] = field(default=None, repr=False)
    smsaRating: Optional[str] = field(default=None, repr=False)
    tvpgRating: Optional[str] = field(default=None, repr=False)
    ytRating: Optional[str] = field(default=None)


@dataclass
class VideoContentDetails(BaseModel):
    """
    A class representing the video content details info.

    Refer: https://developers.google.com/youtube/v3/docs/videos#contentDetails
    """

    duration: Optional[str] = field(default=None)
    dimension: Optional[str] = field(default=None)
    definition: Optional[str] = field(default=None, repr=False)
    caption: Optional[str] = field(default=None, repr=False)
    licensedContent: Optional[bool] = field(default=None, repr=False)
    regionRestriction: Optional[RegionRestriction] = field(default=None, repr=False)
    contentRating: Optional[ContentRating] = field(default=None, repr=False)
    projection: Optional[str] = field(default=None, repr=False)
    hasCustomThumbnail: Optional[bool] = field(default=None, repr=False)

    def get_video_seconds_duration(self):
        if not self.duration:
            return None
        try:
            seconds = isodate.parse_duration(self.duration).total_seconds()
        except ISO8601Error as e:
            raise PyYouTubeException(
                ErrorMessage(status_code=ErrorCode.INVALID_PARAMS, message=e.args[0])
            )
        else:
            return int(seconds)


@dataclass
class VideoTopicDetails(BaseTopicDetails):
    """
    A class representing video's topic detail info.

    Refer: https://developers.google.com/youtube/v3/docs/videos#topicDetails
    """

    # Important:
    # This property has been deprecated as of November 10, 2016.
    # Any topics associated with a video are now returned by the topicDetails.relevantTopicIds[] property value.
    topicIds: Optional[List[str]] = field(default=None, repr=False)
    relevantTopicIds: Optional[List[str]] = field(default=None, repr=False)
    topicCategories: Optional[List[str]] = field(default=None)

    def __post_init__(self):
        """
        If topicIds is not return and relevantTopicIds has return. let relevantTopicIds for topicIds.
        This is for the get_full_topics method.
        :return:
        """
        if self.topicIds is None and self.relevantTopicIds is not None:
            self.topicIds = self.relevantTopicIds


@dataclass
class VideoSnippet(BaseModel, DatetimeTimeMixin):
    """
    A class representing the video snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/videos#snippet
    """

    publishedAt: Optional[str] = field(default=None, repr=False)
    channelId: Optional[str] = field(default=None, repr=False)
    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    thumbnails: Optional[Thumbnails] = field(default=None, repr=False)
    channelTitle: Optional[str] = field(default=None, repr=False)
    tags: Optional[List[str]] = field(default=None, repr=False)
    categoryId: Optional[str] = field(default=None, repr=False)
    liveBroadcastContent: Optional[str] = field(default=None, repr=False)
    defaultLanguage: Optional[str] = field(default=None, repr=False)
    localized: Optional[Localized] = field(default=None, repr=False)
    defaultAudioLanguage: Optional[str] = field(default=None, repr=False)


@dataclass
class VideoStatistics(BaseModel):
    """
    A class representing the video statistics info.

    Refer: https://developers.google.com/youtube/v3/docs/videos#statistics
    """

    viewCount: Optional[int] = field(default=None)
    likeCount: Optional[int] = field(default=None)
    dislikeCount: Optional[int] = field(default=None, repr=False)
    commentCount: Optional[int] = field(default=None, repr=False)


@dataclass
class VideoStatus(BaseModel, DatetimeTimeMixin):
    """
    A class representing the video status info.

    Refer: https://developers.google.com/youtube/v3/docs/videos#status
    """

    uploadStatus: Optional[str] = field(default=None)
    failureReason: Optional[str] = field(default=None, repr=False)
    rejectionReason: Optional[str] = field(default=None, repr=False)
    privacyStatus: Optional[str] = field(default=None)
    publishAt: Optional[str] = field(default=None, repr=False)
    license: Optional[str] = field(default=None, repr=False)
    embeddable: Optional[bool] = field(default=None, repr=False)
    publicStatsViewable: Optional[bool] = field(default=None, repr=False)
    madeForKids: Optional[bool] = field(default=None, repr=False)
    selfDeclaredMadeForKids: Optional[bool] = field(default=None, repr=False)


@dataclass
class VideoLiveStreamingDetails(BaseModel, DatetimeTimeMixin):
    """
    A class representing the video live streaming details.

    Refer: https://developers.google.com/youtube/v3/docs/videos#liveStreamingDetails
    """

    actualStartTime: Optional[str] = field(default=None, repr=False)
    actualEndTime: Optional[str] = field(default=None, repr=False)
    scheduledStartTime: Optional[str] = field(default=None, repr=False)
    scheduledEndTime: Optional[str] = field(default=None, repr=False)
    concurrentViewers: Optional[int] = field(default=None)
    activeLiveChatId: Optional[str] = field(default=None, repr=False)


@dataclass
class Video(BaseResource):
    """
    A class representing the video info.

    Refer: https://developers.google.com/youtube/v3/docs/videos
    """

    snippet: Optional[VideoSnippet] = field(default=None, repr=False)
    contentDetails: Optional[VideoContentDetails] = field(default=None, repr=False)
    status: Optional[VideoStatus] = field(default=None, repr=False)
    statistics: Optional[VideoStatistics] = field(default=None, repr=False)
    topicDetails: Optional[VideoTopicDetails] = field(default=None, repr=False)
    player: Optional[Player] = field(default=None, repr=False)
    liveStreamingDetails: Optional[VideoLiveStreamingDetails] = field(
        default=None, repr=False
    )


@dataclass
class VideoListResponse(BaseApiResponse):
    """
    A class representing the video's retrieve response info.

    Refer: https://developers.google.com/youtube/v3/docs/videos/list#response_1
    """

    items: Optional[List[Video]] = field(default=None, repr=False)


@dataclass
class VideoReportAbuse(BaseModel):
    """
    A class representing the video report abuse body.
    """

    videoId: Optional[str] = field(default=None)
    reasonId: Optional[str] = field(default=None)
    secondaryReasonId: Optional[str] = field(default=None)
    comments: Optional[str] = field(default=None)
    language: Optional[str] = field(default=None)


@dataclass
class VideoRatingItem(BaseModel):
    """
    A class representing the video rating item info.
    """

    videoId: Optional[str] = field(default=None)
    rating: Optional[str] = field(default=None)


@dataclass
class VideoGetRatingResponse(BaseApiResponse):
    """
    A class representing the video rating response.

    References: https://developers.google.com/youtube/v3/docs/videos/getRating#properties
    """

    items: Optional[List[VideoRatingItem]] = field(default=None, repr=False)



"""
    This provide some common utils methods for YouTube resource.
"""
import base64

import isodate
from isodate.isoerror import ISO8601Error

from pyyoutube.error import ErrorMessage, PyYouTubeException

from ytvs import proto


def get_video_duration(duration: str) -> int:
    """
    ISO 8601 형식의 비디오 지속 시간을 초 단위로 변환합니다. 이 형식은 YouTube API에서 비디오 지속 시간을 표현하는 데 사용됩니다.
    예: 'PT14H23M42S'는 '14시간 23분 42초'를 나타냅니다.

    Args:
        duration (str): ISO 8601 형식의 비디오 지속 시간입니다.

    Returns:
        int: 지속 시간을 초 단위로 나타낸 정수값입니다.

    예외 처리:
        ISO8601Error: 지속 시간의 형식이 잘못되었을 때 발생합니다.
    """
    try:
        seconds = isodate.parse_duration(duration).total_seconds()
        return int(seconds)
    except ISO8601Error as e:
        raise PyYouTubeException(
            ErrorMessage(
                status_code=10001,
                message=f"Exception in convert video duration: {duration}. errors: {e}",
            )
        )


def _make_comment_ctoken(video_id, sort=0, offset=0, lc='', secret_key=''):
    """
    YouTube 댓글 페이지의 Continuation Token을 생성.
    :param video_id: 비디오 ID.
    :param sort: 댓글 정렬 방식.
    :param offset: 댓글 페이지의 오프셋.
    :param lc: 로케일 설정.
    :param secret_key: 보안 키.
    :return: 생성된 Continuation Token.
    """
    # 비디오 ID와 보안 키를 바이트로 변환
    video_id = proto.as_bytes(video_id)
    secret_key = proto.as_bytes(secret_key)

    # 페이지 정보 구성
    page_info = proto.string(4, video_id) + proto.uint(6, sort)
    offset_information = proto.nested(4, page_info) + proto.uint(5, offset)
    if secret_key:
        offset_information = proto.string(1, secret_key) + offset_information

    page_params = proto.string(2, video_id)
    if lc:
        page_params += proto.string(6, proto.percent_b64encode(proto.string(15, lc)))

    result = proto.nested(2, page_params) + proto.uint(3, 6) + proto.nested(6, offset_information)
    return base64.urlsafe_b64encode(result).decode('ascii')



import json
import requests

from ytvs.client.unofficial.parameter import DataParameterBuilder
from ytvs.client.unofficial.resources.base_resource import Resource
from ytvs.extractor import YoutubeBasicInfoExtractor
from ytvs.filter import SearchType, Order, VideoDuration, FilterBuilder
from ytvs.utils import _search_attribute, _search_attributes, try_get


class SearchResource(Resource):
    KIND = 'youtube#searchListResponse'   # 검색 결과 응답 유형
    REGION_CODE = 'KR'

    def search(self,
               q: str,
               search_type: SearchType,
               order: Order,
               video_duration: VideoDuration,
               limit=10,
               page_token=None
               ):
        """
        YouTube에서 비디오, 채널, 재생목록을 검색
        :param q: 검색어입니다.
        :param search_type: 검색 유형 (예: 비디오, 채널, 재생목록).
        :param order: 검색 결과의 정렬 순서.
        :param video_duration: 검색할 비디오의 길이.
        :param limit: 검색 결과의 최대 개수.
        :param page_token: 페이징을 위한 토큰.
        :return: 파싱된 검색 결과 데이터.
        """
        # 1. Youtube 검색 데이터 요청
        searched_data = self._request_search(q, search_type, order, video_duration, limit, page_token)
        # 2. 데이터 분석 및 파싱
        parsed_data = self._parse_data(searched_data)
        # 3. 반환
        return parsed_data

    def _request_search(self,
                        q: str,
                        search_type: SearchType,
                        order: Order,
                        video_duration: VideoDuration,
                        limit=10,
                        page_token=None
                        ):
        """
        YouTube에서 주어진 매개변수에 따라 검색 요청을 수행.
        :param q: 검색어.
        :param search_type: 검색할 유형 (예: 비디오, 채널, 재생목록).
        :param order: 검색 결과의 정렬 순서.
        :param video_duration: 검색할 비디오의 길이.
        :param limit: 반환할 결과의 최대 개수.
        :param page_token: 다음 페이지 검색을 위한 토큰.
        :return: 검색된 Raw 데이터.
        """

        has_page_token = True if page_token else False
        if not has_page_token:
            searched_data = self._request_first_search(q, search_type, order, video_duration, limit, page_token)
        else:
            searched_data = self._request_next_page(page_token)

        return searched_data

    def _request_first_search(self,
                              q: str,
                              search_type: SearchType,
                              order: Order,
                              video_duration: VideoDuration,
                              limit=10,
                              page_token=None
                              ):
        """
        첫 페이지의 검색 결과를 요청.
        :param q: 검색어.
        :param search_type: 검색할 유형.
        :param order: 검색 결과의 정렬 순서.
        :param video_duration: 검색할 비디오의 길이.
        :param limit: 반환할 결과의 최대 개수.
        :param page_token: 다음 페이지 검색을 위한 토큰.
        :return: 검색된 첫 페이지의 Raw 데이터.
        """
        filter_builder = FilterBuilder()
        filter_builder.setOrder(order)
        filter_builder.setVideoDuration(video_duration)
        filter_builder.setSearchType(search_type)
        filter = filter_builder.build()

        endpoint = self.build_search_resource_url(q, filter)
        data = json.dumps(DataParameterBuilder().build())
        data = self._client.fetch_url(endpoint, data=data).decode('utf-8')

        # 2. 영상 페이지에서 데이터 추출
        context_data = YoutubeBasicInfoExtractor.extract_yt_initial_data(data)
        return context_data['init_data']

    def _request_next_page(self,
                           continuation_token: str
                           ):
        """
        다음 페이지의 검색 결과를 요청.
        :param continuation_token: 다음 페이지 검색을 위한 토큰.
        :return: 검색된 다음 페이지의 Raw 데이터.
        """
        url = self.build_search_v1_resource_url()
        data = json.dumps(DataParameterBuilder().set_continuation(continuation_token).build())
        response = self._client.fetch_url(url, data=data).decode("utf-8")
        searched_data = json.loads(response)
        return searched_data

    def _parse_data(self, raw_data):
        """
        검색된 Raw 데이터를 파싱하여 필요한 정보를 추출.
        :param raw_data: YouTube 검색 API로부터 받은 Raw 데이터.
        :return: 파싱된 데이터.
        """
        topbar_raw = try_get(raw_data, lambda x: x['topbar']['desktopTopbarRenderer'])
        topbar = YoutubeBasicInfoExtractor.extract_channel_topbar(topbar_raw)
        # 4.
        continuation_renderer = _search_attribute(raw_data, "continuationItemRenderer", max_depth=10)
        if continuation_renderer is not None:
            continuation_command_renderer_raw = try_get(continuation_renderer,
                                                        lambda x: x["continuationEndpoint"]["continuationCommand"],
                                                        dict)
            continuation_command_renderer = YoutubeBasicInfoExtractor.extract_continuation_command(continuation_command_renderer_raw)
        else:
            continuation_command_renderer = {}
        items = []
        item_section_renderer = _search_attributes(raw_data, "itemSectionRenderer", max_depth=10)
        if item_section_renderer is not None:
            video_renderers = _search_attributes(item_section_renderer, "videoRenderer") or []
            # playlist_renderer = _search_attributes(item_section_renderer, "playlistRenderer") or []
            # if playlist_renderer is not None:
            #     playlist = YoutubeBasicInfoExtractor.extract_playlist_renderer(playlist_renderer)
            #     items.append(playlist)
            for video_renderer_raw in video_renderers:
                video = YoutubeBasicInfoExtractor.extract_video_renderer(video_renderer_raw)
                video['kind'] = 'youtube#searchResult'
                items.append(video)

        return {
            'etag': None,
            'kind': self.KIND,
            'regionCode': topbar['country_code'],
            'items': items,
            'nextPageToken': try_get(continuation_command_renderer, lambda x:x['token']),
            'prevPageToken': None,
            'pageInfo': {
                'totalResults': len(items),
                'resultsPerPage': 20,
            }
        }



import requests
import socks, sockshandler
import gzip

from rotate_ip_aws_gw import ApiGateway

from ytvs.exception import HTTPAsymmetricCookieProcessor, FetchError
from ytvs.proxy_manager_pool import ProxyManagerPool
from ytvs.tor_manager import TorManager

try:
    import brotli

    have_brotli = True
except ImportError:
    have_brotli = False
import urllib.parse
import re
import time
import os
import json
import collections
import traceback
import urllib3
from ytvs.constants import DEFAULT_API_KEY

try:
    import brotli

    have_brotli = True
except ImportError:
    have_brotli = False




class UnofficialClient:
    def __init__(self,
                 api_key=DEFAULT_API_KEY,
                 proxy_host=None,
                 tor_port:int = 9050,
                 use_tor: bool = False,
                 aws_apigw_endpoint=None
                 ):
        self.tor_manager = TorManager(tor_port=tor_port)
        self._api_key = api_key
        self._session = requests.Session()
        self._use_tor = use_tor
        self.aws_apigw_endpoint = aws_apigw_endpoint
        if proxy_host is None:
            self._connection_pool = urllib3.PoolManager(cert_reqs='CERT_REQUIRED')
        else:
            self._connection_pool = ProxyManagerPool(proxies=proxy_host)

    def get_pool(self, use_tor: bool):
        # return self._connection_pool
        if not use_tor:
            return self._connection_pool
        return self.tor_manager.get_tor_connection_pool()

    def decode_content(self, content, encoding_header):
        encodings = encoding_header.replace(' ', '').split(',')
        for encoding in reversed(encodings):
            if encoding == 'identity':
                continue
            if encoding == 'br':
                content = brotli.decompress(content)
            elif encoding == 'gzip':
                content = gzip.decompress(content)
        return content

    def fetch_url_response(self,
                           url,
                           headers=(),
                           timeout=15,
                           data=None,
                           cookiejar_send=None,
                           cookiejar_receive=None,
                           use_tor=True,
                           max_redirects=None):
        '''
        returns response, cleanup_function
        When cookiejar_send is set to a CookieJar object,
         those cookies will be sent in the request (but cookies in response will not be merged into it)
        When cookiejar_receive is set to a CookieJar object,
         cookies received in the response will be merged into the object (nothing will be sent from it)
        When both are set to the same object, cookies will be sent from the object,
         and response cookies will be merged into it.
        '''
        headers = dict(headers)  # Note: Calling dict() on a dict will make a copy
        # if have_brotli:
        #     headers['Accept-Encoding'] = 'gzip, br'
        # else:
        #     headers['Accept-Encoding'] = 'gzip'

        # prevent python version being leaked by urllib if User-Agent isn't provided
        #  (urllib will use ex. Python-urllib/3.6 otherwise)
        if 'User-Agent' not in headers and 'user-agent' not in headers and 'User-agent' not in headers:
            headers['User-Agent'] = 'Python-urllib'

        method = "GET"
        if data is not None:
            method = "POST"
            if isinstance(data, str):
                data = data.encode('utf-8')
            elif not isinstance(data, bytes):
                data = urllib.parse.urlencode(data).encode('utf-8')
        if cookiejar_send is not None or cookiejar_receive is not None:  # Use urllib
            req = urllib.request.Request(url, data=data, headers=headers)

            cookie_processor = HTTPAsymmetricCookieProcessor(cookiejar_send=cookiejar_send,
                                                             cookiejar_receive=cookiejar_receive)

            # if use_tor and settings.route_tor:
            #     opener = urllib.request.build_opener(sockshandler.SocksiPyHandler(socks.PROXY_TYPE_SOCKS5, "127.0.0.1", settings.tor_port), cookie_processor)
            # else:
            #     opener = urllib.request.build_opener(cookie_processor)
            opener = urllib.request.build_opener(cookie_processor)

            response = opener.open(req, timeout=timeout)
            cleanup_func = (lambda r: None)

        else:

            if max_redirects:
                retries = urllib3.Retry(3 + max_redirects, redirect=max_redirects, raise_on_redirect=False)
            else:
                retries = urllib3.Retry(3, raise_on_redirect=False)
            pool = self.get_pool(self._use_tor)
            try:
                if self.aws_apigw_endpoint is None:
                    response = pool.request(method, url, headers=headers, body=data,
                                            timeout=timeout,
                                            preload_content=False,
                                            decode_content=False,
                                            retries=retries)
                    response.retries = retries
                else:
                    gw = ApiGateway(site='')
                    response = gw.getAsRequests(url, self.aws_apigw_endpoint)
            except urllib3.exceptions.MaxRetryError as e:
                exception_cause = e.__context__.__context__
                if (isinstance(exception_cause, socks.ProxyConnectionError)
                        and False):
                    msg = ('Failed to connect to Tor. Check that Tor is open and '
                           'that your internet connection is working.\n\n'
                           + str(e))
                    raise FetchError('502', reason='Bad Gateway',
                                     error_message=msg)
                elif isinstance(e.__context__,
                                urllib3.exceptions.NewConnectionError):
                    msg = 'Failed to establish a connection.\n\n' + str(e)
                    raise FetchError(
                        '502', reason='Bad Gateway',
                        error_message=msg)
                else:
                    raise
            cleanup_func = (lambda r: r.release_conn())

        return response, cleanup_func

    def fetch_url(self,
                  url,
                  headers=(),
                  timeout=15,
                  report_text=None,
                  data=None,
                  cookiejar_send=None,
                  cookiejar_receive=None,
                  use_tor=True,
                  debug_name=None):
        while True:
            start_time = time.monotonic()

            # response, cleanup_func = fetch_url_response(
            #     url, headers, timeout=timeout, data=data,
            #     cookiejar_send=cookiejar_send, cookiejar_receive=cookiejar_receive,
            #     use_tor=use_tor)
            response, cleanup_func = self.fetch_url_response(
                url, headers, timeout=timeout, data=data,
                cookiejar_send=cookiejar_send, cookiejar_receive=cookiejar_receive,
                use_tor=self._use_tor)
            response_time = time.monotonic()

            content = response.read()

            read_finish = time.monotonic()

            cleanup_func(response)  # release_connection for urllib3
            content = self.decode_content(content, response.getheader('Content-Encoding', default='identity'))

            if response.status == 429 or (
                response.status == 302 and (response.getheader('Location') == url
                    or response.getheader('Location').startswith(
                           'https://www.google.com/sorry/index'
                       )
                )
            ):
                print(response.status, response.reason, response.getheaders())
                ip = re.search(
                    br'IP address: ((?:[\da-f]*:)+[\da-f]+|(?:\d+\.)+\d+)',
                    content)
                ip = ip.group(1).decode('ascii') if ip else None
                if not ip:
                    ip = re.search(r'IP=((?:\d+\.)+\d+)',
                                   response.getheader('Set-Cookie') or '')
                    ip = ip.group(1) if ip else None

                # don't get new identity if we're not using Tor
                if not use_tor:
                    raise FetchError('429', reason=response.reason, ip=ip)

                print('Error: Youtube blocked the request because the Tor exit node is overutilized. Exit node IP address: %s' % ip)

                # get new identity
                error = self.tor_manager.new_identity(start_time)
                if error:
                    raise FetchError(
                        '429', reason=response.reason, ip=ip,
                        error_message='Automatic circuit change: ' + error)
                else:
                    continue # retry now that we have new identity

            elif response.status >= 400:
                raise FetchError(str(response.status), reason=response.reason,
                                 ip=None)
            break
        if report_text:
            print(report_text, '    Latency:', round(response_time - start_time, 3), '    Read time:',
                  round(read_finish - response_time, 3))
        return content

    def build_continuation_url_and_data(self,
                                        url,
                                        api_key,
                                        continuation_token):
        c_token = continuation_token
        # Caption
        request_url = f'{url}?key={api_key}'
        data = json.dumps({
            'context': {
                'client': {
                    'hl': 'ko',  # en
                    'gl': 'KR',  # US
                    'clientName': 'WEB',  # MWEB
                    'clientVersion': '2.20210804.02.00',
                },
            },
            'params': c_token.replace('=', '%3D'),
        })

        return {'url': url, 'data': data}



"""
    These are i18n language and region related models.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .common import BaseResource, BaseApiResponse


@dataclass
class I18nRegionSnippet(BaseModel):
    """
    A class representing the I18n region snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/i18nRegions#snippet
    """

    gl: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)


@dataclass
class I18nRegion(BaseResource):
    """
    A class representing the I18n region info.

    Refer: https://developers.google.com/youtube/v3/docs/i18nRegions#resource-representation
    """

    snippet: Optional[I18nRegionSnippet] = field(default=None)


@dataclass
class I18nRegionListResponse(BaseApiResponse):
    """
    A class representing the I18n region list response info.

    Refer: https://developers.google.com/youtube/v3/docs/i18nLanguages/list#response_1
    """

    items: Optional[List[I18nRegion]] = field(default=None, repr=False)


@dataclass
class I18nLanguageSnippet(BaseModel):
    """
    A class representing the I18n language snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/i18nLanguages#snippet
    """

    hl: Optional[str] = field(default=None)
    name: Optional[str] = field(default=None)


@dataclass
class I18nLanguage(BaseResource):
    """
    A class representing the I18n language info.

    Refer: https://developers.google.com/youtube/v3/docs/i18nLanguages#resource-representation
    """

    snippet: Optional[I18nLanguageSnippet] = field(default=None)


@dataclass
class I18nLanguageListResponse(BaseApiResponse):
    """
    A class representing the I18n language list response info.

    Refer: https://developers.google.com/youtube/v3/docs/i18nLanguages/list#response_1
    """

    items: Optional[List[I18nLanguage]] = field(default=None, repr=False)



"""
    Captions resource implementation
"""
from typing import Optional, Union

from requests import Response

from pyyoutube.resources.base_resource import Resource
from pyyoutube.media import Media, MediaUpload
from pyyoutube.models import Caption, CaptionListResponse
from pyyoutube.utils.params_checker import enf_comma_separated, enf_parts


class CaptionsResource(Resource):
    """A caption resource represents a YouTube caption track

    References: https://developers.google.com/youtube/v3/docs/captions
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        video_id: Optional[str] = None,
        caption_id: Optional[Union[str, list, tuple, set]] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, CaptionListResponse]:
        """Returns a list of caption tracks that are associated with a specified video.

        Args:
            parts:
                Comma-separated list of one or more caption resource properties.
            video_id:
                The parameter specifies the YouTube video ID of the video for which the API
                should return caption tracks.
            caption_id:
                The id parameter specifies a comma-separated list of IDs that identify the
                caption resources that should be retrieved.
            on_behalf_of_content_owner:
                This parameter can only be used in a properly authorized request.
                Note: This parameter is intended exclusively for YouTube content partners.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Caption data
        """

        params = {
            "part": enf_parts(resource="captions", value=parts),
            "videoId": video_id,
            "id": enf_comma_separated(field="caption_id", value=caption_id),
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }
        response = self._client.request(path="captions", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else CaptionListResponse.from_dict(data)

    def insert(
        self,
        body: Union[dict, Caption],
        media: Media,
        parts: Optional[Union[str, list, tuple, set]] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        sync: Optional[bool] = None,
        **kwargs,
    ) -> MediaUpload:
        """Uploads a caption track.

        Args:
            body:
                Provide caption data in the request body. You can give dataclass or just a dict with data.
            media:
                Caption media data to upload.
            parts:
                The part parameter specifies the caption resource parts that
                the API response will include. Set the parameter value to snippet.
            on_behalf_of_content_owner:
                This parameter can only be used in a properly authorized request.
                Note: This parameter is intended exclusively for YouTube content partners.
            sync:
                The sync parameter indicates whether YouTube should automatically synchronize the caption
                file with the audio track of the video.
                If you set the value to true, YouTube will disregard any time codes that are in the uploaded
                caption file and generate new time codes for the captions.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Caption data.
        """

        params = {
            "part": enf_parts(resource="captions", value=parts),
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "sync": sync,
            **kwargs,
        }
        # Build a media upload instance.
        media_upload = MediaUpload(
            client=self._client,
            resource="captions",
            media=media,
            params=params,
            body=body.to_dict_ignore_none(),
        )
        return media_upload

    def update(
        self,
        body: Union[dict, Caption],
        media: Optional[Media] = None,
        parts: Optional[Union[str, list, tuple, set]] = None,
        on_behalf_of_content_owner: Optional[str] = None,
        sync: Optional[bool] = None,
        return_json: bool = False,
        **kwargs,
    ) -> Union[dict, Caption, MediaUpload]:
        """Updates a caption track.

        Args:
            body:
                Provide caption data in the request body. You can give dataclass or just a dict with data.
            media:
                New caption media.
            parts:
                The part parameter specifies the caption resource parts that
                the API response will include. Set the parameter value to snippet.
            on_behalf_of_content_owner:
                This parameter can only be used in a properly authorized request.
                Note: This parameter is intended exclusively for YouTube content partners.
            sync:
                The sync parameter indicates whether YouTube should automatically synchronize the caption
                file with the audio track of the video.
                If you set the value to true, YouTube will disregard any time codes that are in the uploaded
                caption file and generate new time codes for the captions.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.
        Returns:
            Caption data.

        """
        params = {
            "part": enf_parts(resource="captions", value=parts),
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "sync": sync,
            **kwargs,
        }
        if media is not None:
            # Build a media upload instance.
            media_upload = MediaUpload(
                client=self._client,
                resource="captions",
                media=media,
                params=params,
                body=body.to_dict_ignore_none(),
            )
            return media_upload

        response = self._client.request(
            method="PUT",
            path="captions",
            params=params,
            json=body,
        )
        data = self._client.parse_response(response=response)
        return data if return_json else Caption.from_dict(data)

    def download(
        self,
        caption_id: str,
        on_behalf_of_content_owner: Optional[str] = None,
        tfmt: Optional[str] = None,
        tlang: Optional[str] = None,
        **kwargs,
    ) -> Response:
        """Downloads a caption track.

        Args:
            caption_id:
                ID for the caption track that is being deleted.
            on_behalf_of_content_owner:
                This parameter can only be used in a properly authorized request.
                Note: This parameter is intended exclusively for YouTube content partners.
            tfmt:
                Specifies that the caption track should be returned in a specific format.
                Supported values are:
                    sbv – SubViewer subtitle
                    scc – Scenarist Closed Caption format
                    srt – SubRip subtitle
                    ttml – Timed Text Markup Language caption
                    vtt – Web Video Text Tracks caption
            tlang:
                Specifies that the API response should return a translation of the specified caption track.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Response form YouTube.
        """
        params = {
            "id": caption_id,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            "tfmt": tfmt,
            "tlang": tlang,
            **kwargs,
        }
        response = self._client.request(
            path=f"captions/{caption_id}",
            params=params,
        )
        return response

    def delete(
        self,
        caption_id: str,
        on_behalf_of_content_owner: Optional[str] = None,
        **kwargs,
    ) -> bool:
        """Deletes a specified caption track.

        Args:
            caption_id:
                ID for the caption track that is being deleted.
            on_behalf_of_content_owner:
                This parameter can only be used in a properly authorized request.
                Note: This parameter is intended exclusively for YouTube content partners.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Delete status

        Raises:
            PyYouTubeException: Request not success.
        """
        params = {
            "id": caption_id,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }

        response = self._client.request(path="captions", method="DELETE", params=params)
        if response.ok:
            return True
        self._client.parse_response(response=response)



"""
    These are category related models.
    Include VideoCategory
"""
from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .common import BaseApiResponse, BaseResource


@dataclass
class CategorySnippet(BaseModel):
    """
    This is base category snippet for video and guide.
    """

    channelId: Optional[str] = field(default=None)
    title: Optional[str] = field(default=None)


@dataclass
class VideoCategorySnippet(CategorySnippet):
    """
    A class representing video category snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/videoCategories#snippet
    """

    assignable: Optional[bool] = field(default=None, repr=False)


@dataclass
class VideoCategory(BaseResource):
    """
    A class representing video category info.

    Refer: https://developers.google.com/youtube/v3/docs/videoCategories
    """

    snippet: Optional[VideoCategorySnippet] = field(default=None, repr=False)


@dataclass
class VideoCategoryListResponse(BaseApiResponse):
    """
     A class representing the video category's retrieve response info.

    Refer: https://developers.google.com/youtube/v3/docs/videoCategories/list#response_1
    """

    items: Optional[List[VideoCategory]] = field(default=None, repr=False)



"""
    These are watermark related models.
"""

from dataclasses import dataclass, field
from typing import Optional

from .base import BaseModel


@dataclass
class WatermarkTiming(BaseModel):
    type: Optional[str] = field(default=None)
    offsetMs: Optional[int] = field(default=None, repr=False)
    durationMs: Optional[int] = field(default=None, repr=False)


@dataclass
class WatermarkPosition(BaseModel):
    type: Optional[str] = field(default=None)
    cornerPosition: Optional[str] = field(default=None, repr=False)


@dataclass
class Watermark(BaseModel):
    """
    A class representing the watermark info.

    References: https://developers.google.com/youtube/v3/docs/watermarks#resource-representation
    """

    timing: Optional[WatermarkTiming] = field(default=None, repr=False)
    position: Optional[WatermarkPosition] = field(default=None, repr=False)
    imageUrl: Optional[str] = field(default=None)
    imageBytes: Optional[bytes] = field(default=None, repr=False)
    targetChannelId: Optional[str] = field(default=None, repr=False)



"""
    i18n language resource implementation.
"""

from typing import Optional, Union

from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import I18nLanguageListResponse
from pyyoutube.utils.params_checker import enf_parts


class I18nLanguagesResource(Resource):
    """An i18nLanguage resource identifies an application language that the YouTube website supports.
    The application language can also be referred to as a UI language

    References: https://developers.google.com/youtube/v3/docs/i18nLanguages
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        hl: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, I18nLanguageListResponse]:
        """Returns a list of application languages that the YouTube website supports.

        Args:
            parts:
                Comma-separated list of one or more i18n languages resource properties.
                Accepted values: snippet.
            hl:
                Specifies the language that should be used for text values in the API response.
                The default value is en_US.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            i18n language data
        """
        params = {
            "part": enf_parts(resource="i18nLanguages", value=parts),
            "hl": hl,
            **kwargs,
        }
        response = self._client.request(path="i18nLanguages", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else I18nLanguageListResponse.from_dict(data)



import inspect

from ytvs.client.unofficial import resources
from ytvs.client.unofficial.client import UnofficialClient
from ytvs.client.unofficial.resources.base_resource import Resource
from ytvs.constants import DEFAULT_API_KEY


def is_resource_endpoint(obj):
    """
    객체가 Resource의 인스턴스인지 확인.
    :param obj: 검사할 객체입니다.
    :return: 객체가 Resource의 인스턴스인 경우 True를 반환.
    """
    return isinstance(obj, Resource)


class UnofficialApi:
    # 비공식 APi를 구성하는 Resource 클래스들의 인스턴스를 초기화.
    search = resources.SearchResource()
    video = resources.VideoResource()
    channel = resources.ChannelResource()
    comment = resources.CommentResource()
    comment_thread = resources.CommentThreadResource()
    transcript = resources.TranscriptResource()

    def __new__(cls,
                proxies=None,
                use_tor: bool = False,
                tor_port:int = 9050,
                aws_apigw_endpoint=None,
                *args, **kwargs):
        """
        unofficialapi 클래스의 새 인스턴스를 생성.
        모든 sub resource들을 unofficialclient와 연결.
        :return: unofficialapi의 새 인스턴스.
        """
        self = super().__new__(cls)

        # 비공식 클라이언트 인스턴스를 생성.
        client = UnofficialClient(proxy_host=proxies,
                                  use_tor=use_tor,
                                  tor_port=tor_port,
                                  aws_apigw_endpoint=aws_apigw_endpoint)

        # 모든 리소스 엔드포인트를 UnofficialClient와 연결합니다.
        sub_resources = inspect.getmembers(self, is_resource_endpoint)
        for name, resource in sub_resources:
            resource_cls = type(resource)
            resource = resource_cls(client)
            setattr(self, name, resource)

        return self

    def __init__(self,
                 api_key=None,
                 proxies=None,
                 use_tor:bool=False,
                 aws_apigw_endpoint=None,
                 tor_port:int=9050
                 ):
        """
        UnofficialApi 클래스의 생성자.
        :param api_key: API 키입니다. 기본값은 ytvs.constants에서 정의된 DEFAULT_API_KEY.
        """
        self.api_key = api_key or DEFAULT_API_KEY
        self.proxies = proxies
        self.use_tor = use_tor
        self.tor_port = tor_port
        self.aws_apigw_endpoint = aws_apigw_endpoint



"""
    These are channel related models.

    References: https://developers.google.com/youtube/v3/docs/channels#properties
"""
from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .common import (
    BaseResource,
    BaseTopicDetails,
    Thumbnails,
    BaseApiResponse,
    Localized,
)
from .mixins import DatetimeTimeMixin


@dataclass
class RelatedPlaylists(BaseModel):
    """
    A class representing the channel's related playlists info

    References: https://developers.google.com/youtube/v3/docs/channels#contentDetails.relatedPlaylists
    """

    likes: Optional[str] = field(default=None, repr=False)
    uploads: Optional[str] = field(default=None)


@dataclass
class ChannelBrandingSettingChannel(BaseModel):
    """
    A class representing the channel branding setting's channel info.

    References: https://developers.google.com/youtube/v3/docs/channels#brandingSettings.channel
    """

    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    keywords: Optional[str] = field(default=None, repr=False)
    trackingAnalyticsAccountId: Optional[str] = field(default=None, repr=False)
    moderateComments: Optional[bool] = field(default=None, repr=False)
    unsubscribedTrailer: Optional[str] = field(default=None, repr=False)
    defaultLanguage: Optional[str] = field(default=None, repr=False)
    country: Optional[str] = field(default=None, repr=False)


@dataclass
class ChannelBrandingSettingImage(BaseModel):
    """
    A class representing the channel branding setting's image info.

    References: https://developers.google.com/youtube/v3/docs/channels#brandingSettings.image
    """

    bannerExternalUrl: Optional[str] = field(default=None, repr=False)


@dataclass
class ChannelSnippet(BaseModel, DatetimeTimeMixin):
    """
    A class representing the channel snippet info.

    References: https://developers.google.com/youtube/v3/docs/channels#snippet
    """

    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    customUrl: Optional[str] = field(default=None, repr=False)
    publishedAt: Optional[str] = field(default=None, repr=False)
    thumbnails: Optional[Thumbnails] = field(default=None, repr=False)
    defaultLanguage: Optional[str] = field(default=None, repr=False)
    localized: Optional[Localized] = field(default=None, repr=False)
    country: Optional[str] = field(default=None, repr=False)


@dataclass
class ChannelContentDetails(BaseModel):
    """
    A class representing the channel's content info.

    References: https://developers.google.com/youtube/v3/docs/channels#contentDetails
    """

    relatedPlaylists: Optional[RelatedPlaylists] = field(default=None)


@dataclass
class ChannelStatistics(BaseModel):
    """
    A class representing the Channel's statistics info.

    References: https://developers.google.com/youtube/v3/docs/channels#statistics
    """

    viewCount: Optional[int] = field(default=None)
    subscriberCount: Optional[int] = field(default=None)
    hiddenSubscriberCount: Optional[bool] = field(default=None, repr=False)
    videoCount: Optional[int] = field(default=None, repr=False)


@dataclass
class ChannelTopicDetails(BaseTopicDetails):
    """
    A class representing the channel's topic detail info.

    References: https://developers.google.com/youtube/v3/docs/channels#topicDetails
    """

    # Important:
    # topicIds maybe has deprecated.
    # see more: https://developers.google.com/youtube/v3/revision_history#november-10-2016
    topicIds: Optional[List[str]] = field(default=None, repr=False)
    topicCategories: Optional[List[str]] = field(default=None)


@dataclass
class ChannelStatus(BaseModel):
    """
    A class representing the channel's status info.

    References: https://developers.google.com/youtube/v3/docs/channels#status
    """

    privacyStatus: Optional[str] = field(default=None)
    isLinked: Optional[bool] = field(default=None, repr=False)
    longUploadsStatus: Optional[str] = field(default=None, repr=False)
    madeForKids: Optional[bool] = field(default=None, repr=False)
    selfDeclaredMadeForKids: Optional[bool] = field(default=None, repr=False)


@dataclass
class ChannelBrandingSetting(BaseModel):
    """
    A class representing the channel branding settings info.

    References: https://developers.google.com/youtube/v3/docs/channels#brandingSettings
    """

    channel: Optional[ChannelBrandingSettingChannel] = field(default=None)
    image: Optional[ChannelBrandingSettingImage] = field(default=None)


@dataclass
class ChannelAuditDetails(BaseModel):
    """A class representing the channel audit details info.

    References: https://developers.google.com/youtube/v3/docs/channels#auditDetails
    """

    overallGoodStanding: Optional[bool] = field(default=None)
    communityGuidelinesGoodStanding: Optional[bool] = field(default=None, repr=True)
    copyrightStrikesGoodStanding: Optional[bool] = field(default=None, repr=True)
    contentIdClaimsGoodStanding: Optional[bool] = field(default=None, repr=True)


@dataclass
class ChannelContentOwnerDetails(BaseModel):
    """A class representing the channel data relevant for YouTube Partners.

    References: https://developers.google.com/youtube/v3/docs/channels#contentOwnerDetails
    """

    contentOwner: Optional[str] = field(default=None)
    timeLinked: Optional[str] = field(default=None)


@dataclass
class Channel(BaseResource):
    """
    A class representing the channel's info.

    References: https://developers.google.com/youtube/v3/docs/channels
    """

    snippet: Optional[ChannelSnippet] = field(default=None, repr=False)
    contentDetails: Optional[ChannelContentDetails] = field(default=None, repr=False)
    statistics: Optional[ChannelStatistics] = field(default=None, repr=False)
    topicDetails: Optional[ChannelTopicDetails] = field(default=None, repr=False)
    status: Optional[ChannelStatus] = field(default=None, repr=False)
    brandingSettings: Optional[ChannelBrandingSetting] = field(default=None, repr=False)
    auditDetails: Optional[ChannelAuditDetails] = field(default=None, repr=False)
    contentOwnerDetails: Optional[ChannelContentOwnerDetails] = field(
        default=None, repr=False
    )
    localizations: Optional[dict] = field(default=None, repr=False)


@dataclass
class ChannelListResponse(BaseApiResponse):
    """
    A class representing the channel's retrieve response info.

    References: https://developers.google.com/youtube/v3/docs/channels/list#response
    """

    items: Optional[List[Channel]] = field(default=None, repr=False)



"""
    These are common models for multi resource.
"""
from dataclasses import dataclass, field
from typing import Optional, List

from .base import BaseModel


@dataclass
class Thumbnail(BaseModel):
    """
    A class representing the thumbnail resource info.

    Refer: https://developers.google.com/youtube/v3/docs/channels#snippet.thumbnails.(key).url
    """

    url: Optional[str] = field(default=None)
    width: Optional[int] = field(default=None, repr=False)
    height: Optional[int] = field(default=None, repr=False)


@dataclass
class Thumbnails(BaseModel):
    """
    A class representing the multi thumbnail resource info.

    Refer: https://developers.google.com/youtube/v3/docs/channels#snippet.thumbnails
    """

    default: Optional[Thumbnail] = field(default=None)
    medium: Optional[Thumbnail] = field(default=None, repr=False)
    high: Optional[Thumbnail] = field(default=None, repr=False)
    standard: Optional[Thumbnail] = field(default=None, repr=False)
    maxres: Optional[Thumbnail] = field(default=None, repr=False)


@dataclass
class Topic(BaseModel):
    """
    A class representing the channel topic info. this model also suitable for video.

    Refer:
        https://developers.google.com/youtube/v3/docs/channels#topicDetails.topicIds[]
        https://developers.google.com/youtube/v3/docs/videos#topicDetails.topicIds[]

    This model is customized for parsing topic id. YouTube Data Api not return this.
    """

    id: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)


@dataclass
class BaseTopicDetails(BaseModel):
    """
    This is the base model for channel or video topic details.
    """

    topicIds: List[str] = field(default=None, repr=False)

    def get_full_topics(self):
        """
        Convert topicIds list to Topic model list
        :return: List[Topic]
        """
        from pyyoutube import TOPICS

        r: List[Topic] = []
        if self.topicIds:
            for topic_id in self.topicIds:
                topic = Topic.from_dict(
                    {"id": topic_id, "description": TOPICS.get(topic_id)}
                )
                r.append(topic)
        return r


@dataclass
class Localized(BaseModel):
    """
    A class representing the channel or video snippet localized info.

    Refer:
        https://developers.google.com/youtube/v3/docs/channels#snippet.localized
        https://developers.google.com/youtube/v3/docs/videos#snippet.localized
    """

    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None, repr=False)


@dataclass
class PageInfo(BaseModel):
    """
    This is data model for save paging data.
    Note:
        totalResults is only an approximation/estimate.
        Refer:
            https://stackoverflow.com/questions/43507281/totalresults-count-doesnt-match-with-the-actual-results-returned-in-youtube-v3
    """

    totalResults: Optional[int] = field(default=None)
    resultsPerPage: Optional[int] = field(default=None)


@dataclass
class BaseApiResponse(BaseModel):
    """
    This is Data Api response structure when retrieve data.
    They both have same response structure, but items.

    Refer:
        https://developers.google.com/youtube/v3/docs/channels/list#response_1
        https://developers.google.com/youtube/v3/docs/playlistItems/list#response_1
    """

    kind: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None, repr=False)
    nextPageToken: Optional[str] = field(default=None, repr=False)
    prevPageToken: Optional[str] = field(default=None, repr=False)
    pageInfo: Optional[PageInfo] = field(default=None, repr=False)


@dataclass
class BaseResource(BaseModel):
    """
    This is a base model for different resource type.

    Refer: https://developers.google.com/youtube/v3/docs#resource-types
    """

    kind: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None, repr=False)
    id: Optional[str] = field(default=None)


@dataclass
class ResourceId(BaseModel):
    """
    A class representing the subscription snippet resource info.
    Refer:
        1. https://developers.google.com/youtube/v3/docs/playlistItems#snippet.resourceId
        2. https://developers.google.com/youtube/v3/docs/subscriptions#snippet.resourceId
        3. https://developers.google.com/youtube/v3/docs/activities#contentDetails.social.resourceId
    """

    kind: Optional[str] = field(default=None)
    videoId: Optional[str] = field(default=None)
    channelId: Optional[str] = field(default=None)
    playlistId: Optional[str] = field(default=None)


@dataclass
class Player(BaseModel):
    """
    A class representing the video,playlist player info.

    Refer:
        https://developers.google.com/youtube/v3/docs/videos#player

    """

    embedHtml: Optional[str] = field(default=None)
    # Important:
    # follows attributions maybe not exists.
    embedHeight: Optional[int] = field(default=None, repr=False)
    embedWidth: Optional[int] = field(default=None, repr=False)



from typing import Optional, List
from ytvs.client.official.api import OfficialApi
from ytvs.client.unofficial.api import UnofficialApi
from ytvs.filter import SearchType, Order, VideoDuration
from ytvs.models import CommentThreadListResponse, SearchListResponse, ChannelListResponse, VideoListResponse, \
    CommentListResponse, CaptionListResponse
from ytvs.models.transcript import TranscriptSnippet


class YoutubeApi(object):
    """
       YouTubeApi 클래스는 YouTube의 공식 및 비공식 API에 대한 접근을 제공.
       이 클래스는 다양한 YouTube 리소스에 대한 검색, 비디오 정보 조회, 채널 정보 조회, 댓글 스레드 조회 등의 기능을 제공.
       사용자는 공식 API와 비공식 API 중 하나를 선택하여 사용 가능.
   """

    def __init__(self,
                 *,
                 api_keys: List[str],
                 use_tor: bool = False,
                 tor_port:int = 9050,
                 aws_apigw_endpoint=None,
                 proxies=None
                 ):
        """
        YouTube API 클래스의 생성자. 공식 및 비공식 API 클라이언트를 초기화.
        :param api_key: YouTube API를 사용하기 위한 API 키.
        """
        self._official_client = OfficialApi(api_keys=api_keys, proxies=proxies)
        #self._unofficial_client = UnofficialApi(proxies=proxies)
        self._unofficial_client = UnofficialApi(use_tor=use_tor,
                                                tor_port=tor_port,
                                                aws_apigw_endpoint=aws_apigw_endpoint,
                                                proxies=proxies)
        self._use_official = False

    def search(self,
               *,
               q: str,
               search_type: SearchType,
               order: Order,
               video_duration: VideoDuration,
               limit=10,
               page_token: Optional[str] = None,
               use_official: Optional[bool] = False
               ) -> SearchListResponse:
        """
        YouTube에서 키워드를 이용해 검색.
        지원:공식API   -    O
            비공식API  -    O
        :param q: 검색어.
        :param search_type: 검색 유형.
        :param order: 검색 결과의 정렬 순서.
        :param video_duration: 검색할 비디오의 길이.
        :param limit: 검색 결과의 최대 개수.
        :param page_token: 페이징을 위한 토큰입
        :param use_official: 공식 API를 사용할지 여부.
        :return: 검색 결과.
        """
        if use_official:
            # 공식 API를 사용하여 채널 정보를 가져옵니다.
            return self._official_client.search_by_keywords(
                q=q,
                search_type=search_type,
                order=order,
                video_duration=video_duration,
                limit=limit,
                page_token=page_token
            )
        else:
            # 비공식 API를 사용하여 채널 정보를 가져옵니다.
            result_dict = self._unofficial_client.search.search(
                q,
                search_type,
                order,
                video_duration,
                limit,
                page_token
            )
            return SearchListResponse.from_dict(result_dict)

    def get_channel_info(self,
                         *,
                         channel_id: str,
                         use_official: Optional[bool] = False
                         ) -> ChannelListResponse:
        """
        지정된 YouTube 채널의 정보를 조회. 공식 API와 비공식 API를 선택적으로 사용 가능.
        :param channel_id: 정보를 가져올 채널의 ID.
        :param use_official: 공식 API를 사용할지 여부. 기본값은 False.
        :return: 채널 정보를 포함한 ChannelListResponse 객체.
        """
        if use_official:
            # 공식 API를 사용하여 채널 정보를 가져옵니다.
            return self._official_client.get_channel_info(channel_id=channel_id)
        else:
            # 비공식 API를 사용하여 채널 정보를 가져옵니다.
            result_dict = self._unofficial_client.channel.get_channel_info(channel_id=channel_id)
            return ChannelListResponse.from_dict(result_dict)

    def get_video_by_id(self,
                        *,
                        video_id,
                        use_official=False
                        ) -> VideoListResponse:
        """
        YouTube 채널 정보를 획득.
        지원:공식API    -    O
            비공식API   -    O
        :param channel_id: 정보를 가져올 채널의 ID.
        :param use_official: 공식 API를 사용할지 여부.
        :return: 채널 정보입니다.
        """
        if use_official:
            # 공식 API를 사용하여 채널 정보를 가져옵니다.
            return self._official_client.get_video_by_id(video_id=video_id)
        else:
            # 비공식 API를 사용하여 채널 정보를 가져옵니다.
            data = self._unofficial_client.video.get_video_by_id(video_id=video_id)
            return VideoListResponse.from_dict(data)

    def get_comment_threads(self,
                            *,
                            video_id,
                            use_official=False) -> CommentThreadListResponse:
        """
        YouTube 비디오의 댓글 스레드를 조회.
        지원:공식API   -    O
            비공식API  -    O
        :param video_id: 댓글 스레드를 가져올 비디오의 ID입니다.
        :param use_official: 공식 API를 사용할지 여부입니다.
        :return: 댓글 스레드 정보입니다.
        """
        if use_official:
            # 공식 API를 사용하여 채널 정보를 가져옵니다.
            return self._official_client.get_comment_threads(video_id=video_id)
        else:
            # 비공식 API를 사용하여 채널 정보를 가져옵니다.
            data = self._unofficial_client.comment_thread.get_comment_threads(video_id=video_id)
            return CommentThreadListResponse.from_dict(data)

    def get_comments_official(self,
                              *,
                              parent_id: str,
                              page_token: str
                              ) -> CommentListResponse:
        """

        지원:
            공식API   -    O
            비공식API -    O
        :return:
        """
        # 공식 API를 사용하여 채널 정보를 가져옵니다.
        return self._official_client.get_comments(parent_id=parent_id, page_token=page_token)

    def get_comments_unofficial(self,
                                *,
                                page_token: str) -> CommentListResponse:
        """

        지원:
            공식API   -    O
            비공식API -    O
        :return:
        """
        # 비공식 API를 사용하여 채널 정보를 가져옵니다.
        data = self._unofficial_client.comment.get_comments(page_token=page_token)
        return CommentListResponse.from_dict(data)

    def get_transcript_by_video_id(self,
                                   *
                                   video_id: str,
                                   use_official: bool = False
                                   ) -> TranscriptSnippet:
        """
        YouTube 비디오의 자막 정보를 조회.
        지원:공식API   -    X
            비공식API  -    O
        :param video_id: 자막 정보를 가져올 비디오의 ID입니다.
        :param use_official: 공식 API를 사용할지 여부입니다.
        :return: 자막 정보입니다. 공식 API에서는 지원되지 않습니다.
        """
        if use_official:
            # 공식 API는 지원하지 않음.
            raise NotImplemented("Not supported in official api")
        else:
            # 비공식 API를 사용하여 채널 정보를 가져옵니다.
            data = self._unofficial_client.transcript.get_transcript_by_video_id(video_id=video_id)
            return TranscriptSnippet.from_dict(data)

    def get_caption_by_video_id(self,
                                *
                                video_id: str,
                                use_official: bool = False
                                ) -> CaptionListResponse:
        """
        YouTube 비디오의 캡션 정보를 조회.
        지원:공식API   -    O
            비공식API  -    X
        :param video_id: 캡션 정보를 가져올 비디오의 ID.
        :param use_official: 공식 API를 사용할지 여부.
        :return: 캡션 정보입니다. 비공식 API는 미지원.
        """
        if use_official:
            # 공식 API를 사용하여 채널 정보를 가져옵니다.
            return self._official_client.get_captions_by_video(video_id=video_id)
        else:
            # 비공식 API는 지원하지 않음.
            raise NotImplemented("Not supported in unofficial api")



"""
  These are subscription related models.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .common import BaseApiResponse, BaseResource, ResourceId, Thumbnails
from .mixins import DatetimeTimeMixin


@dataclass
class SubscriptionSnippet(BaseModel, DatetimeTimeMixin):
    """
    A class representing the subscription snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/subscriptions#snippet
    """

    publishedAt: Optional[str] = field(default=None, repr=False)
    channelTitle: Optional[str] = field(default=None, repr=False)
    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    resourceId: Optional[ResourceId] = field(default=None, repr=False)
    channelId: Optional[str] = field(default=None, repr=False)
    thumbnails: Optional[Thumbnails] = field(default=None, repr=False)


@dataclass
class SubscriptionContentDetails(BaseModel):
    """
    A class representing the subscription contentDetails info.

    Refer: https://developers.google.com/youtube/v3/docs/subscriptions#contentDetails
    """

    totalItemCount: Optional[int] = field(default=None)
    newItemCount: Optional[int] = field(default=None)
    activityType: Optional[str] = field(default=None, repr=False)


@dataclass
class SubscriptionSubscriberSnippet(BaseModel):
    """
    A class representing the subscription subscriberSnippet info.

    Refer: https://developers.google.com/youtube/v3/docs/subscriptions#subscriberSnippet
    """

    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    channelId: Optional[str] = field(default=None, repr=False)
    thumbnails: Optional[Thumbnails] = field(default=None, repr=False)


@dataclass
class Subscription(BaseResource):
    """
    A class representing the subscription info.

    Refer: https://developers.google.com/youtube/v3/docs/subscriptions
    """

    snippet: Optional[SubscriptionSnippet] = field(default=None)
    contentDetails: Optional[SubscriptionContentDetails] = field(
        default=None, repr=False
    )
    subscriberSnippet: Optional[SubscriptionSubscriberSnippet] = field(
        default=None, repr=False
    )


@dataclass
class SubscriptionListResponse(BaseApiResponse):
    """
    A class representing the subscription's retrieve response info.

    Refer: https://developers.google.com/youtube/v3/docs/subscriptions/list#response_1
    """

    items: Optional[List[Subscription]] = field(default=None, repr=False)



from dataclasses import dataclass
from typing import Optional


class ErrorCode:
    HTTP_ERROR = 10000
    MISSING_PARAMS = 10001
    INVALID_PARAMS = 10002
    NEED_AUTHORIZATION = 10003
    AUTHORIZE_URL_FIRST = 10004


@dataclass
class ErrorMessage:
    status_code: Optional[int] = None
    message: Optional[str] = None




import re

import requests
import json

from ytvs.utils import try_get, str_to_int, _search_regex, extract_view_count, extract_like_count, \
    convert_relative_to_absolute_date, _search_attribute, select_first_item_in_list, number_str_abbr_to_int, \
    is_korean_number, _search_attributes


class YoutubeBasicInfoExtractor:
    @staticmethod
    def _extract_channel_renderer(renderer):
        channel_id = try_get(renderer, lambda x: x['channelId'], str)
        type_name = 'channel'
        title = try_get(renderer, lambda x: x['title']['simpleText'], str)
        thumbnail = try_get(renderer, lambda x: x['thumbnail'], str)
        return {
            "id": channel_id,
            "type": type_name,
            "title": title,
            "thumbnail": thumbnail,
        }

    @staticmethod
    def extract_playlist_renderer(renderer):
        playlist_id = try_get(renderer, lambda x: x['channelId'], str)
        type_name = 'channel'
        thumbnails = try_get(renderer, lambda x: x['thumbnails'], str)
        title = try_get(renderer, lambda x: x["title"]["simpleText"], str)
        length = try_get(renderer, lambda x: x['videoCount'], str)
        videos = try_get(renderer, lambda x: x['videos'], str)
        video_count = try_get(renderer, lambda x: x['videoCount'], str)
        is_live = False
        return {
            "id": playlist_id,
            "type": type_name,
            "thumbnail": thumbnails,
            "title": title,
            "length": length,
            "videos": videos,
            "videoCount": video_count,
            "isLive": is_live
        }

    @staticmethod
    def _check_if_is_live_video(renderer):
        badges = try_get(renderer, lambda x: x['badges'])
        badge_metadata_renderer = _search_attribute(badges, 'metadataBadgeRenderer')
        badge_style = try_get(badge_metadata_renderer, lambda x: x['style'], str)
        if badge_style == "BADGE_STYLE_TYPE_LIVE_NOW":
            return True

        thumbnail_overlay = try_get(renderer, lambda x: x['thumbnailOverlays'])
        thumbnail_overlay_time_status_renderers = _search_attributes(thumbnail_overlay,
                                                                     'thumbnailOverlayTimeStatusRenderer') or []
        for thumbnail_overlay_time_status_renderer in thumbnail_overlay_time_status_renderers:
            style = try_get(thumbnail_overlay_time_status_renderer, "style")
            if style == 'LIVE':
                return True

        return False

    @staticmethod
    def extract_video_renderer(renderer):
        video_id = try_get(renderer, lambda x: x['videoId'], str)
        thumbnails = try_get(renderer, lambda x: x['thumbnail']['thumbnails'], list)
        thumbnail_urls = {}
        for idx, thumbnail in enumerate(thumbnails):

            if idx == 0:
                thumbnail_urls['default'] = thumbnail
            elif idx == 1:
                thumbnail_urls['medium'] = thumbnail
            elif idx == 2:
                thumbnail_urls['high'] = thumbnail

        owner_text_renderer = try_get(renderer, lambda x: x['ownerText'], dict)
        browse_endpoint_raw = _search_attribute(owner_text_renderer, 'browseEndpoint', max_depth=5)
        browse_endpoint = YoutubeBasicInfoExtractor.extract_browse_endpoint(browse_endpoint_raw)
        channel_handle = browse_endpoint['channel_url']
        channel_id = browse_endpoint['channel_id']

        title_runs = try_get(renderer, lambda x: x['title']['runs'], list)
        titles = []
        for title_run in title_runs:
            title_text = title_run['text']
            titles.append(title_text)
        title = "".join(titles)

        description_runs = try_get(renderer, lambda x: x['detailedMetadataSnippets'][0]['snippetText']['runs'],
                                   list) or []
        descriptions = []
        for description_run in description_runs:
            description_text = description_run['text']
            descriptions.append(description_text)
        description = "\r\n".join(descriptions)

        published_time_text = try_get(renderer, lambda x: x['publishedTimeText']['simpleText'], str)
        if not published_time_text is None:
            published_at = convert_relative_to_absolute_date(published_time_text)
        else:
            published_at = None

        duration_text = try_get(renderer, lambda x: x['lengthText']['simpleText'])

        channel_title_runs = try_get(renderer, lambda x: x['ownerText']['runs'], list)
        channel_titles = []
        for channel_title_run in channel_title_runs:
            channel_title = channel_title_run['text']
            channel_titles.append(channel_title)
        channel_title = "".join(channel_titles)
        is_live = YoutubeBasicInfoExtractor._check_if_is_live_video(renderer)

        return {
            'id': {
                'channelId': channel_id,
                'kind': 'youtube#video',
                'playlistId': None,
                'videoId': video_id
            },
            'etag': None,
            'kind': 'youtube#searchResult',
            'snippet': {
                "title": title,
                'channelId': channel_id,
                'channelTitle': channel_title,
                'description': description,
                "duration": duration_text,
                'liveBroadcastContent': is_live,
                'publishedAt': published_at,
                'thumbnails': thumbnail_urls,
            }
        }

    @staticmethod
    def _extract_browse_endpoint(renderer):
        channel_id = try_get(renderer, lambda x: x['browseId'], str)
        channel_handle = try_get(renderer, lambda x: x['canonicalBaseUrl'], str)
        return {
            'channel_id': channel_id,
            'channel_handle': channel_handle
        }

    @staticmethod
    def extract_metadata(metadata):
        title = try_get(metadata, lambda x: x['title'], str)
        description = try_get(metadata, lambda x: x['description'], str)
        rss_url = try_get(metadata, lambda x: x['rssUrl'], str)
        channel_conversion_url = try_get(metadata, lambda x: x['channelConversionUrl'], str)
        keywords = try_get(metadata, lambda x: x['keywords'], list)
        avatar_urls = try_get(metadata, lambda x: x['avatar']['thumbnails'], list)
        channel_url = try_get(metadata, lambda x: x['channelUrl'], list)
        is_family_safe = try_get(metadata, lambda x: x['isFamilySafe'], str)
        available_country_codes = try_get(metadata, lambda x: x['availableCountryCodes'], str)
        vanity_channel_url = try_get(metadata, lambda x: x['vanityChannelUrl'], str)

        return {
            'title': title,
            'description': description,
            'rss_url': rss_url,
            'channel_conversion_url': channel_conversion_url,
            'keywords': keywords,
            'avatar_urls': avatar_urls,
            'channel_url': channel_url,
            'is_family_safe': is_family_safe,
            'available_country_codes': available_country_codes,
            'vanity_channel_url': vanity_channel_url
        }

    @staticmethod
    def extract_channel_topbar(topbar):
        country_code = try_get(topbar, lambda x: x['countryCode'], str)
        return {"country_code": country_code}

    @staticmethod
    def _extract_thumbnail_renderer(renderer):
        if renderer is None:
            return None
        thumbnail_urls = {}
        for idx, banner_url in enumerate(renderer):
            if idx == 0:
                thumbnail_urls['default'] = banner_url
            elif idx == 1:
                thumbnail_urls['medium'] = banner_url
            elif idx == 2:
                thumbnail_urls['high'] = banner_url
        return thumbnail_urls

    @staticmethod
    def extract_channel_header(header):
        channel_id = try_get(header, lambda x: x['channelId'], str)
        title = try_get(header, lambda x: x['title'], str)
        avatars_renderer = try_get(header, lambda x: x['avatar']['thumbnails'], list)
        avatar_urls = YoutubeBasicInfoExtractor._extract_thumbnail_renderer(avatars_renderer)
        banners_renderer = try_get(header, lambda x: x['banner']['thumbnails'], list)
        banner_urls = YoutubeBasicInfoExtractor._extract_thumbnail_renderer(banners_renderer)

        tv_banners_renderer = try_get(header, lambda x: x['tvBanner']['thumbnails'], list)
        tv_banner_urls = YoutubeBasicInfoExtractor._extract_thumbnail_renderer(tv_banners_renderer)
        mobile_banners_renderer = try_get(header, lambda x: x['mobileBanner']['thumbnails'], list)
        mobile_banner_urls = YoutubeBasicInfoExtractor._extract_thumbnail_renderer(mobile_banners_renderer)
        channel_handle_text = try_get(header, lambda x: x['channelHandleText']['runs'][0]['text'])
        video_count_text = try_get(header, lambda x: x['videosCountText']['runs'][1]['text'])
        video_count = str_to_int(video_count_text)
        subscriber_count_text = try_get(header, lambda x: x['subscriberCountText']['simpleText'], str)
        if is_korean_number(subscriber_count_text):
            subscriber_count = number_str_abbr_to_int(subscriber_count_text)
        else:
            subscriber_count = str_to_int(subscriber_count_text)

        return {
            'channel_id': channel_id,
            'title': title,
            'avatar_urls': avatar_urls,
            'banner_urls': banner_urls,
            'tv_banner_urls': tv_banner_urls,
            'mobile_banner_urls': mobile_banner_urls,
            'channel_handle_text': channel_handle_text,
            'video_count': video_count,
            'subscriber_count': subscriber_count
        }

    @staticmethod
    def _extract_transcript_segment_renderer(renderer):
        target_id = try_get(renderer, lambda x: x['targetId'], str)
        start_ms = try_get(renderer, lambda x: x['startMs'], str)
        end_ms = try_get(renderer, lambda x: x['endMs'], str)
        tracking_params = try_get(renderer, lambda x: x['trackingParams'], str)
        text_runs = try_get(renderer, lambda x: x['snippet']['runs'], list)
        texts = []
        for text_run in text_runs:
            texts.append(text_run['text'])
        text = "".join(texts)

        return {
            'target_id': target_id,
            'start_ms': start_ms,
            'end_ms': end_ms,
            'text': text,
            'tracking_params': tracking_params
        }

    @staticmethod
    def _extract_transcript_section_header_renderer(renderer):
        start_ms = try_get(renderer, lambda x: x['startMs'], str)
        end_ms = try_get(renderer, lambda x: x['endMs'], str)
        title = try_get(renderer, lambda x: x['accessibility']['accessibilityData']['label'], str)
        tracking_params = try_get(renderer, lambda x: x['trackingParams'], str)

        return {
            "start_ms": start_ms,
            "end_ms": end_ms,
            "title": title,
            "tracking_params": tracking_params
        }

    @staticmethod
    def extract_transcript_segment_list_renderer(renderer):
        initial_segments = try_get(renderer, lambda x: x['initialSegments'])
        transcripts = []
        for initial_segment in initial_segments:
            for keyname in initial_segment.keys():
                if keyname == "transcriptSectionHeaderRenderer":
                    transcript = YoutubeBasicInfoExtractor._extract_transcript_section_header_renderer(initial_segment[keyname])
                    transcripts.append(transcript)
                elif keyname == 'transcriptSegmentRenderer':
                    transcript = YoutubeBasicInfoExtractor._extract_transcript_segment_renderer(initial_segment[keyname])
                    transcripts.append(transcript)

        return transcripts

    @staticmethod
    def extract_content_text(renderer):
        content_text_raws = try_get(renderer, lambda x: x['contentText']['runs'], list)
        content_texts = []
        for content_text_raw in content_text_raws:
            content_texts.append(content_text_raw['text'])
        content_text = "".join(content_texts)
        return content_text


    @staticmethod
    def extract_watch_endpoint(renderer):
        video_id = try_get(renderer, lambda x: x['videoId'], str)
        params = try_get(renderer, lambda x: x['params'], str)

        return {
            'videoId': video_id,
            'params': params
        }

    @staticmethod
    def _extract_web_command_metadata(renderer):
        send_post = try_get(renderer, lambda x: x['sendPost'], str)
        api_url = try_get(renderer, lambda x: x['apiUrl'], str)

        return {
            'sendPost': send_post,
            'apiUrl': api_url
        }

    @staticmethod
    def extract_continuation_command(renderer):
        token = try_get(renderer, lambda x: x['token'], str)
        request = try_get(renderer, lambda x: x['request'], str)

        return {
            'token': token,
            'request': request
        }

    @staticmethod
    def extract_continuation_item_renderer(renderer):
        click_tracking_params = try_get(renderer, lambda x: x['clickTrackingParams'], str)
        command_metadata_raw = try_get(renderer, lambda x: x['commandMetadata'], dict)
        web_command_metadata_raw = try_get(command_metadata_raw, lambda x: x['web_command_metadata'], dict)
        web_command_metadata = YoutubeBasicInfoExtractor._extract_web_command_metadata(web_command_metadata_raw)
        continuation_command_raw = try_get(renderer, lambda x: x['continuationCommand'], dict)
        continuation_comamnd = YoutubeBasicInfoExtractor.extract_continuation_command(continuation_command_raw)

        return {
            'click_tracking_params': click_tracking_params,
            'web_command_metadata': web_command_metadata,
            'continuation_command': continuation_comamnd,
        }

    @staticmethod
    def extract_comments_headerer_renderer(renderer):
        title = try_get(renderer, lambda x: x['titleText']['runs'][0]['text'], str)
        count = try_get(renderer, lambda x: x['commentsCount']['runs'][0]['text'], str)

        return {
            'title': title,
            'count': count
        }

    @staticmethod
    def _extract_button_renderer(renderer):
        text_runs = try_get(renderer, lambda x: x['text']['runs'], list)
        texts = []
        for text_run in text_runs:
            text = text_run['text']
            texts.append(text)
        text = "".join(texts)
        style = try_get(renderer, lambda x: x['style'], str)

        return {
            'text': text,
            'style': style
        }

    @staticmethod
    def extract_toggle_button_renderer(renderer):
        title = try_get(renderer, lambda x: x['defaultTooltip'], str)
        accessibility_label = try_get(renderer, lambda x: x['accessibilityData']['accessibilityData']['label'], str)
        count = extract_like_count(accessibility_label)
        return {
            'title': title,
            'accessibility_label': accessibility_label,
            'count': count
        }

    @staticmethod
    def extract_comment_replies_renderer(renderer):
        continuation_endpoint = try_get(renderer,
                                        lambda x: x['contents'][0]['continuationItemRenderer'][
                                            'continuationEndpoint'], dict)
        comment_replies_continuation = YoutubeBasicInfoExtractor.extract_continuation_item_renderer(continuation_endpoint)
        continuation_token = try_get(comment_replies_continuation, lambda x: x['continuation_command']['token'], str)
        return {
            'nextPageToken': continuation_token,
        }

    @staticmethod
    def extract_comment_renderer(renderer):
        author_text = try_get(renderer, (lambda x: x['authorText']['runs'][0]['text'],
                                         lambda x: x['authorText']['simpleText']), str)
        thumbnails = try_get(renderer, lambda x: x['authorThumbnail']['thumbnails'], list)
        thumbnail = select_first_item_in_list(thumbnails)



        browse_endpoint_raw = try_get(renderer, lambda x: x['authorEndpoint']['browseEndpoint'], dict)
        browse_endpoint = YoutubeBasicInfoExtractor.extract_browse_endpoint(browse_endpoint_raw)
        author_channel_url = browse_endpoint['channel_url']
        author_channel_id = browse_endpoint['channel_id']

        content_text = YoutubeBasicInfoExtractor.extract_content_text(renderer)
        # content_texts = []
        # for content_text_raw in content_text_raws:
        #     content_texts.append(content_text_raw['text'])
        # content_text = "".join(content_texts)

        comment_id = try_get(renderer, lambda x: x['commentId'], str)
        published_time_text = try_get(renderer, lambda x: x['publishedTimeText']['runs'][0]['text'], str)
        published_at = convert_relative_to_absolute_date(published_time_text)

        #
        action_buttons = try_get(renderer, lambda x: x['actionButtons']['commentActionButtonsRenderer'], dict)
        like_button = _search_attribute(action_buttons, 'likeButton')
        like_button_info = YoutubeBasicInfoExtractor.extract_toggle_button_renderer(like_button['toggleButtonRenderer'])

        watch_endpoint_renderer = try_get(renderer,
                                          lambda x: x['publishedTimeText']['runs'][0]['navigationEndpoint'][
                                              'watchEndpoint'], dict)
        watch_endpoint = YoutubeBasicInfoExtractor.extract_watch_endpoint(watch_endpoint_renderer)
        video_id = watch_endpoint['videoId']

        # reply_button = _search_attribute(action_buttons, 'replyButton')
        # reply_button_info = _extract_button_renderer(reply_button)

        return {  # root
            'etag': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
            'id': comment_id,
            'kind': 'youtube#commentThread',
            'replies': None,
            # snippet
            'snippet': {
                'canReply': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                'channelId': browse_endpoint['channel_id'],  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                'isPublic': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                # TopLevelComment
                'topLevelComment': {
                    'etag': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    'id': comment_id,
                    'kind': 'youtube#comment',
                    'snippet': {
                        'authorChannelUrl': author_channel_url,
                        'authorDisplayName': author_text,
                        'authorProfileImageUrl': thumbnail,
                        'canRate': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                        'channelId': author_channel_id,
                        'videoId':video_id,
                        'likeCount': like_button_info.get('count'),
                        'moderationStatus': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                        'parentId': '',
                        'publishedAt': published_at,
                        'textDisplay': content_text,
                        'textOriginal': content_text,
                        'updatedAt': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                        'viewerRating': 'none',  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    },
                    'totalReplyCount': 0,
                    'videoId': id,

                }
            }
        }



    @staticmethod
    def extract_comment_thread_renderer(renderer):
        author_text = try_get(renderer, (lambda x: x['authorText']['runs'][0]['text'],
                                         lambda x: x['authorText']['simpleText']), str)
        thumbnails = try_get(renderer, lambda x: x['authorThumbnail']['thumbnails'], list)
        thumbnail = select_first_item_in_list(thumbnails)



        browse_endpoint_raw = try_get(renderer, lambda x: x['authorEndpoint']['browseEndpoint'], dict)
        browse_endpoint = YoutubeBasicInfoExtractor.extract_browse_endpoint(browse_endpoint_raw)
        author_channel_url = browse_endpoint['channel_url']
        author_channel_id = browse_endpoint['channel_id']

        content_text = YoutubeBasicInfoExtractor.extract_content_text(renderer)
        # content_texts = []
        # for content_text_raw in content_text_raws:
        #     content_texts.append(content_text_raw['text'])
        # content_text = "".join(content_texts)

        comment_id = try_get(renderer, lambda x: x['commentId'], str)
        published_time_text = try_get(renderer, lambda x: x['publishedTimeText']['runs'][0]['text'], str)
        published_at = convert_relative_to_absolute_date(published_time_text)

        #
        action_buttons = try_get(renderer, lambda x: x['actionButtons']['commentActionButtonsRenderer'], dict)
        like_button = _search_attribute(action_buttons, 'likeButton')
        like_button_info = YoutubeBasicInfoExtractor.extract_toggle_button_renderer(like_button['toggleButtonRenderer'])

        watch_endpoint_renderer = try_get(renderer,
                                          lambda x: x['publishedTimeText']['runs'][0]['navigationEndpoint'][
                                              'watchEndpoint'], dict)
        watch_endpoint = YoutubeBasicInfoExtractor.extract_watch_endpoint(watch_endpoint_renderer)
        video_id = watch_endpoint['videoId']

        # reply_button = _search_attribute(action_buttons, 'replyButton')
        # reply_button_info = _extract_button_renderer(reply_button)

        return {  # root
            'etag': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
            'id': comment_id,
            'kind': 'youtube#commentThread',
            'replies': None,
            # snippet
            'snippet': {
                'canReply': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                'channelId': browse_endpoint['channel_id'],  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                'isPublic': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                # TopLevelComment
                'topLevelComment': {
                    'etag': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    'id': comment_id,
                    'kind': 'youtube#comment',
                    'snippet': {
                        'authorChannelUrl': author_channel_url,
                        'authorDisplayName': author_text,
                        'authorProfileImageUrl': thumbnail,
                        'canRate': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                        'channelId': author_channel_id,
                        'videoId':video_id,
                        'likeCount': like_button_info.get('count'),
                        'moderationStatus': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                        'parentId': '',
                        'publishedAt': published_at,
                        'textDisplay': content_text,
                        'textOriginal': content_text,
                        'updatedAt': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                        'viewerRating': 'none',  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    },
                    'totalReplyCount': 0,
                    'videoId': id,

                }
            }
        }



    @staticmethod
    def extract_comment_renderer(renderer):
        author_text = try_get(renderer, (lambda x: x['authorText']['runs'][0]['text'],
                                         lambda x: x['authorText']['simpleText']), str)
        thumbnails = try_get(renderer, lambda x: x['authorThumbnail']['thumbnails'], list)
        thumbnail = select_first_item_in_list(thumbnails)



        browse_endpoint_raw = try_get(renderer, lambda x: x['authorEndpoint']['browseEndpoint'], dict)
        browse_endpoint = YoutubeBasicInfoExtractor.extract_browse_endpoint(browse_endpoint_raw)
        author_channel_url = browse_endpoint['channel_url']
        author_channel_id = browse_endpoint['channel_id']

        content_text = YoutubeBasicInfoExtractor.extract_content_text(renderer)
        # content_texts = []
        # for content_text_raw in content_text_raws:
        #     content_texts.append(content_text_raw['text'])
        # content_text = "".join(content_texts)

        comment_id = try_get(renderer, lambda x: x['commentId'], str)
        published_time_text = try_get(renderer, lambda x: x['publishedTimeText']['runs'][0]['text'], str)
        published_at = convert_relative_to_absolute_date(published_time_text)

        #
        action_buttons = try_get(renderer, lambda x: x['actionButtons']['commentActionButtonsRenderer'], dict)
        like_button = _search_attribute(action_buttons, 'likeButton')
        like_button_info = YoutubeBasicInfoExtractor.extract_toggle_button_renderer(like_button['toggleButtonRenderer'])

        watch_endpoint_renderer = try_get(renderer,
                                          lambda x: x['publishedTimeText']['runs'][0]['navigationEndpoint'][
                                              'watchEndpoint'], dict)
        watch_endpoint = YoutubeBasicInfoExtractor.extract_watch_endpoint(watch_endpoint_renderer)
        video_id = watch_endpoint['videoId']

        # reply_button = _search_attribute(action_buttons, 'replyButton')
        # reply_button_info = _extract_button_renderer(reply_button)

        return {  # root

                'etag': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                'id': comment_id,
                'kind': 'youtube#comment',
                'snippet': {
                    'authorChannelUrl': author_channel_url,
                    'authorDisplayName': author_text,
                    'authorProfileImageUrl': thumbnail,
                    'canRate': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    'channelId': author_channel_id,
                    'videoId':video_id,
                    'likeCount': like_button_info.get('count'),
                    'moderationStatus': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    'parentId': '',
                    'publishedAt': published_at,
                    'textDisplay': content_text,
                    'textOriginal': content_text,
                    'updatedAt': None,  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                    'viewerRating': 'none',  # @TODO: 이정보가 필요하다면, 획득하는 방법에 대해 먼저 조사가 필요
                },
                'totalReplyCount': 0,
                'videoId': id,

            }

    @staticmethod
    def extract_browse_endpoint(renderer):
        channel_id = try_get(renderer, lambda x: x['browseId'], str)
        channel_url = try_get(renderer, lambda x: x['canonicalBaseUrl'], str)
        return {
            'channel_id': channel_id,
            'channel_url': channel_url
        }

    @staticmethod
    def extract_comments_header_renderer(renderer):
        if renderer is None:
            return None
        title_runs = try_get(renderer, lambda x: x['headerText']['runs'])
        titles = []
        for title_run in title_runs:
            titles.append(title_run['text'])
        title = "".join(titles)
        comment_count = try_get(renderer, lambda x: x['commentCount']['simpleText'])
        if comment_count is None:
            comment_count = "0"

        if is_korean_number(comment_count):
            comment_count = number_str_abbr_to_int(comment_count)
        else:
            comment_count = str_to_int(comment_count)
        return {
            'title': title,
            'comment_count': comment_count,
        }

    @staticmethod
    def extract_video_owner_renderer(renderer):
        title = try_get(renderer, lambda x: x['title']['runs'][0]['text'])
        channel_handle = try_get(renderer, lambda x: x['title']['runs'][0]['navigationEndpoint']['browseEndpoint'][
            'canonicalBaseUrl'])
        channel_id = try_get(renderer,
                             lambda x: x['title']['runs'][0]['navigationEndpoint']['browseEndpoint']['browseId'])
        thumbnails = try_get(renderer, lambda x: x['thumbnail']['thumbnails'], dict)
        # TODO: 정규표현식을 이용해서 파싱해야됨
        subscribers = try_get(renderer, lambda x: x['subscriberCountText']['simpleText'], str)

        return {
            'title': title,
            'channel_handle': channel_handle,
            'channel_id': channel_id,
            "thumbnails": thumbnails,
            'subscribers': subscribers
        }

    @staticmethod
    def extract_video(renderer):
        # video_id = renderer['videoId']
        # 영상 제목 획득
        # titles = [try_get(renderer, lambda x: x['title']['runs'][0]['text'], str),
        #           try_get(renderer, lambda x: x['title']['runs'][1]['text'], str),
        #           try_get(renderer, lambda x: x['title']['runs'][2]['text'], str)]
        # title = "".join(filter(lambda x: x is not None, titles))
        title_runs = try_get(renderer, lambda x: x['title']['runs'], list)
        titles = []
        for title_run in title_runs:
            title_text = title_run['text']
            titles.append(title_text)
        title = "".join(titles)

        # 영상 설명 획득
        view_count_text = try_get(
            renderer, (lambda x: x['viewCount']['videoViewCountRenderer']['viewCount']['simpleText'],
                       lambda x: x['viewCount']['videoViewCountRenderer']['originalViewCount'], # Live streaming
                       ), str
        )
        # 해시태그 목록 획득
        super_title_links_raw = try_get(renderer, lambda x: x['superTitleLink']['runs']) or []
        super_title_links = []
        for super_title_link_raw in super_title_links_raw:
            title_link = super_title_link_raw['text']
            if title_link.strip() == "":
                continue
            super_title_links.append(title_link)

        # 조회수 횟수 획득
        view_count = extract_view_count(view_count_text)
        # view_count = str_to_int(_search_regex(r'^([\d,]+)', re.sub(r'\s', '', view_count_text), 'view count', default=None))

        # Like 횟수 획득
        menu_renderer = try_get(renderer, lambda x: x['videoActions']['menuRenderer'], dict)
        like_button_view_model = try_get(menu_renderer,
                                         lambda x: x['topLevelButtons'][0]['segmentedLikeDislikeButtonViewModel'][
                                             'likeButtonViewModel']['likeButtonViewModel']['toggleButtonViewModel'][
                                             'toggleButtonViewModel'][
                                             'defaultButtonViewModel'][
                                             'buttonViewModel'], dict)

        like_count_text = try_get(like_button_view_model, lambda x: x['accessibilityText'])
        like_count = extract_like_count(like_count_text)

        published_at = try_get(renderer, lambda x: x['dateText']['simpleText'])
        if  published_at is not None:
            published_at = convert_relative_to_absolute_date(published_at)
        else:
            published_at = None

        return {
            'title': title,
            'view_count': view_count,
            'like_count': like_count,
            'super_title_links': super_title_links,
            'published_at': published_at
        }

    @staticmethod
    def extract_yt_initial_data(html_text: str):
        yt_init_data = html_text.split("var ytInitialData =")
        if yt_init_data and len(yt_init_data) > 1:
            data = yt_init_data[1].split("</script>")[0].strip()[:-1]

            if "innertubeApiKey" in html_text:
                api_token = html_text.split("innertubeApiKey")[1].strip().split(",")[0].split('"')[2]
            else:
                api_token = None

            if "INNERTUBE_CONTEXT" in html_text:
                context = json.loads(html_text.split("INNERTUBE_CONTEXT")[1].strip()[2:-2])
            else:
                context = None

            init_data = json.loads(data)
            return {"init_data": init_data, "api_token": api_token, "context": context}
        else:
            print("cannot_get_init_data")
            return None

    @staticmethod
    def get_youtube_init_data(self, url):
        init_data = {}
        api_token = None
        context = None
        try:
            page = requests.get(url)
            yt_init_data = page.text.split("var ytInitialData =")
            if yt_init_data and len(yt_init_data) > 1:
                data = yt_init_data[1].split("</script>")[0].strip()[:-1]

                if "innertubeApiKey" in page.text:
                    api_token = page.text.split("innertubeApiKey")[1].strip().split(",")[0].split('"')[2]

                if "INNERTUBE_CONTEXT" in page.text:
                    context = json.loads(page.text.split("INNERTUBE_CONTEXT")[1].strip()[2:-2])

                init_data = json.loads(data)
                return {"init_data": init_data, "api_token": api_token, "context": context}
            else:
                print("cannot_get_init_data")
                return None
        except Exception as ex:
            print(ex)



import base64
import json

from ytvs import proto
from ytvs.client.unofficial.parameter import DataParameterBuilder
from ytvs.client.unofficial.resources.base_resource import Resource
from ytvs.constants import MOBILE_XHR_HEADERS, CONTENT_TYPE_JSON_HEADER
from ytvs.extractor import YoutubeBasicInfoExtractor
from ytvs.utils import _search_attribute, _search_attributes, try_get
from ytvs.youtube_utils import _make_comment_ctoken


class CommentThreadResource(Resource):
    KIND = 'youtube#commentThreadListResponse',# 댓글 스레드 응답 유형

    def get_comment_threads(self,
                            video_id: str,
                            sort=0,
                            offset=0,
                            lc='',
                            secret_key='',
                            page_token=None
                            ):
        """
        YouTube 비디오의 댓글 스레드를 조회.
        :param video_id: 비디오 ID.
        :param sort: 댓글 정렬 방식.
        :param offset: 댓글 페이지의 오프셋.
        :param lc: 로케일 설정.
        :param secret_key: 보안 키.
        :param page_token: 페이지 토큰.
        :return: 파싱된 댓글 스레드 데이터.
        """
        # 1. 댓글 데이터 요청
        comment_raw = self._request_data(video_id,
                                         sort,
                                         offset,
                                         lc,
                                         secret_key,
                                         page_token
                                         )

        # 2. 획득한 댓글 Raw 데이터에서 필요한 정보만 파싱
        comment = self._parse_comment_data(comment_raw)

        # 3. 추가 데이터 병합
        comment_items = comment['items']
        for comment_item in comment_items:
            comment_item['videoId'] = video_id

        # 4. 반환
        return comment

    def _request_data(self,
                      video_id: str,
                      sort=0,
                      offset=0,
                      lc='',
                      secret_key='',
                      page_token=None
                      ):
        """
        YouTube API를 통해 댓글 데이터를 요청하는 함수.
        :param video_id: 댓글을 요청할 비디오의 ID.
        :param sort: 댓글 정렬 방식.
        :param offset: 댓글 페이지의 오프셋.
        :param lc: 로케일 설정.
        :param secret_key: 보안 키.
        :param page_token: 페이지 토큰.
        :return: YouTube API로부터 받은 Raw 댓글 데이터.
        """

        if page_token is None:
            c_token = _make_comment_ctoken(video_id, sort=1 - sort, lc=lc)
        else:
            c_token = page_token
        endpoint = self.build_continuous_resource_url()
        data = json.dumps(DataParameterBuilder().set_continuation(c_token).build())

        content = self._client.fetch_url(
            endpoint, headers=MOBILE_XHR_HEADERS + CONTENT_TYPE_JSON_HEADER, data=data)
        content = content.decode('utf-8')

        raw_json = json.loads(content)
        return raw_json

    def _parse_comment_data(self, raw_data):
        """
         YouTube API로부터 받은 Raw 댓글 데이터를 파싱하는 함수.
         :param raw_data: YouTube API로부터 받은 Raw 댓글 데이터.
         :return: 파싱된 댓글 데이터.
         """
        # 댓글 Header 정보 획득
        header_renderer = _search_attribute(raw_data, 'commentsHeaderRenderer', max_depth=10)
        header = YoutubeBasicInfoExtractor.extract_comments_headerer_renderer(header_renderer)

        # 댓글 본문 정보 획득
        comments = []
        comment_thread_renderers = _search_attributes(raw_data, 'commentThreadRenderer')
        if comment_thread_renderers is None:
            total_comment = 0
            continuation_token = None
        else:
            total_comment = len(comment_thread_renderers)

            for comment_thread_renderer in comment_thread_renderers:
                # Comment thread
                comment_renderer_raw = try_get(comment_thread_renderer, lambda x: x['comment']['commentRenderer'], dict)
                comment = YoutubeBasicInfoExtractor.extract_comment_thread_renderer(comment_renderer_raw)
                # Comment replies
                comment_replies_renderer_raw = try_get(comment_thread_renderer,
                                                       lambda x: x['replies']['commentRepliesRenderer'], dict)
                comment_replies = YoutubeBasicInfoExtractor.extract_comment_replies_renderer(comment_replies_renderer_raw)
                comment['replies'] = comment_replies
                # comment['videoId'] = video_id
                comments.append(comment)

            # 댓글 페이징 정보 획득
            comment_continuation_renderer = _search_attributes(raw_data, 'continuationItemRenderer')
            comment_continuation = YoutubeBasicInfoExtractor.extract_continuation_item_renderer(
                comment_continuation_renderer['continuationEndpoint'])
            continuation_token = try_get(comment_continuation, lambda x: x['continuation_command']['token'], str)

        return {
            'kind': self.KIND,
            'etag': None,
            'items': comments,
            'nextPageToken': continuation_token,
            'pageInfo': {
                'resultsPerPage': 20,
                'totalResults': total_comment,
            },
            'prevPageToken': None
        }



from dataclasses import dataclass, asdict
from typing import Type, TypeVar

from dataclasses_json import DataClassJsonMixin
from dataclasses_json.core import Json, _decode_dataclass

A = TypeVar("A", bound="DataClassJsonMixin")


@dataclass
class BaseModel(DataClassJsonMixin):
    """Base model class for instance use."""

    @classmethod
    def from_dict(cls: Type[A], kvs: Json, *, infer_missing=False) -> A:
        # save original data for lookup
        cls._json = kvs
        return _decode_dataclass(cls, kvs, infer_missing)

    def to_dict_ignore_none(self):
        return asdict(
            obj=self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )



"""
    Thumbnails resources implementation.
"""
from typing import Optional

from pyyoutube.resources.base_resource import Resource
from pyyoutube.media import Media, MediaUpload


class ThumbnailsResource(Resource):
    """A thumbnail resource identifies different thumbnail image sizes associated with a resource.

    References: https://developers.google.com/youtube/v3/docs/thumbnails
    """

    def set(
        self,
        video_id: str,
        media: Media,
        on_behalf_of_content_owner: Optional[str] = None,
        **kwargs: Optional[dict],
    ) -> MediaUpload:
        params = {
            "videoId": video_id,
            "onBehalfOfContentOwner": on_behalf_of_content_owner,
            **kwargs,
        }
        # Build a media upload instance.
        media_upload = MediaUpload(
            client=self._client,
            resource="thumbnails/set",
            media=media,
            params=params,
        )
        return media_upload



import urllib.parse
from typing import Optional, Union
from requests import Response
from ytvs.error import ErrorMessage


class HTTPAsymmetricCookieProcessor(urllib.request.BaseHandler):
    '''Separate cookiejars for receiving and sending'''

    def __init__(self, cookiejar_send=None, cookiejar_receive=None):
        import http.cookiejar
        self.cookiejar_send = cookiejar_send
        self.cookiejar_receive = cookiejar_receive

    def http_request(self, request):
        if self.cookiejar_send is not None:
            self.cookiejar_send.add_cookie_header(request)
        return request

    def http_response(self, request, response):
        if self.cookiejar_receive is not None:
            self.cookiejar_receive.extract_cookies(response, request)
        return response

    https_request = http_request
    https_response = http_response


class FetchError(Exception):
    def __init__(self, code, reason='', ip=None, error_message=None):
        if error_message:
            string = code + ' ' + reason + ': ' + error_message
        else:
            string = 'HTTP error during request: ' + code + ' ' + reason
        Exception.__init__(self, string)
        self.code = code
        self.reason = reason
        self.ip = ip
        self.error_message = error_message


class PyYouTubeException(Exception):
    """
    This is a return demo:
    {'error': {'errors': [{'domain': 'youtube.parameter',
    'reason': 'missingRequiredParameter',
    'message': 'No filter selected. Expected one of: forUsername, managedByMe, categoryId, mine, mySubscribers, id, idParam',
    'locationType': 'parameter',
    'location': ''}],
    'code': 400,
    'message': 'No filter selected. Expected one of: forUsername, managedByMe, categoryId, mine, mySubscribers, id, idParam'}}
    """

    def __init__(self, response: Optional[Union[ErrorMessage, Response]]):
        self.status_code: Optional[int] = None
        self.error_type: Optional[str] = None
        self.message: Optional[str] = None
        self.response: Optional[Union[ErrorMessage, Response]] = response
        self.error_handler()

    def error_handler(self):
        """
        Error has two big type(but not the error type.): This module's error, Api return error.
        So This will change two error to one format
        """
        if isinstance(self.response, ErrorMessage):
            self.status_code = self.response.status_code
            self.message = self.response.message
            self.error_type = "PyYouTubeException"
        elif isinstance(self.response, Response):
            res_data = self.response.json()
            if "error" in res_data:
                error = res_data["error"]
                if isinstance(error, dict):
                    self.status_code = res_data["error"]["code"]
                    self.message = res_data["error"]["message"]
                else:
                    self.status_code = self.response.status_code
                    self.message = error
                self.error_type = "YouTubeException"

    def __repr__(self):
        return (
            f"{self.error_type}(status_code={self.status_code},message={self.message})"
        )

    def __str__(self):
        return self.__repr__()



"""
    These are playlistItem related models.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .mixins import DatetimeTimeMixin
from .common import BaseApiResponse, BaseResource, ResourceId, Thumbnails


@dataclass
class PlaylistItemContentDetails(BaseModel, DatetimeTimeMixin):
    """
    A class representing the playlist item's content details info.

    Refer: https://developers.google.com/youtube/v3/docs/playlistItems#contentDetails
    """

    videoId: Optional[str] = field(default=None)
    note: Optional[str] = field(default=None, repr=False)
    videoPublishedAt: Optional[str] = field(default=None)
    startAt: Optional[str] = field(default=None, repr=False)
    endAt: Optional[str] = field(default=None, repr=False)


@dataclass
class PlaylistItemSnippet(BaseModel, DatetimeTimeMixin):
    """
    A class representing the playlist item's snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/playlistItems#snippet
    """

    publishedAt: Optional[str] = field(default=None, repr=False)
    channelId: Optional[str] = field(default=None, repr=False)
    title: Optional[str] = field(default=None)
    description: Optional[str] = field(default=None)
    thumbnails: Optional[Thumbnails] = field(default=None, repr=False)
    channelTitle: Optional[str] = field(default=None, repr=False)
    videoOwnerChannelTitle: Optional[str] = field(default=None, repr=False)
    videoOwnerChannelId: Optional[str] = field(default=None, repr=False)
    playlistId: Optional[str] = field(default=None, repr=False)
    position: Optional[int] = field(default=None, repr=False)
    resourceId: Optional[ResourceId] = field(default=None, repr=False)


@dataclass
class PlaylistItemStatus(BaseModel):
    """
    A class representing the playlist item's status info.

    Refer: https://developers.google.com/youtube/v3/docs/playlistItems#status
    """

    privacyStatus: Optional[str] = field(default=None)


@dataclass
class PlaylistItem(BaseResource):
    """
    A class representing the playlist item's info.

    Refer: https://developers.google.com/youtube/v3/docs/playlistItems
    """

    snippet: Optional[PlaylistItemSnippet] = field(default=None, repr=False)
    contentDetails: Optional[PlaylistItemContentDetails] = field(
        default=None, repr=False
    )
    status: Optional[PlaylistItemStatus] = field(default=None, repr=False)


@dataclass
class PlaylistItemListResponse(BaseApiResponse):
    """
    A class representing the playlist item's retrieve response info.

    Refer: https://developers.google.com/youtube/v3/docs/playlistItems/list#response_1
    """

    items: Optional[List[PlaylistItem]] = field(default=None, repr=False)



"""
    Comment resource implementation.
"""

from typing import Optional, Union

from pyyoutube.error import PyYouTubeException, ErrorMessage, ErrorCode
from pyyoutube.resources.base_resource import Resource
from pyyoutube.models import Comment, CommentListResponse
from pyyoutube.utils.params_checker import enf_comma_separated, enf_parts


class CommentsResource(Resource):
    """A comment resource contains information about a single YouTube comment.

    References: https://developers.google.com/youtube/v3/docs/comments
    """

    def list(
        self,
        parts: Optional[Union[str, list, tuple, set]] = None,
        comment_id: Optional[Union[str, list, tuple, set]] = None,
        parent_id: Optional[str] = None,
        max_results: Optional[int] = None,
        text_format: Optional[str] = None,
        page_token: Optional[str] = None,
        return_json: bool = False,
        **kwargs: Optional[dict],
    ) -> Union[dict, CommentListResponse]:
        """Returns a list of comments that match the API request parameters.

        Args:
            parts:
                Comma-separated list of one or more comment resource properties.
            comment_id:
                Specifies a comma-separated list of comment IDs for the resources that are being retrieved.
            parent_id:
                Specifies the ID of the comment for which replies should be retrieved.
            max_results:
                The parameter specifies the maximum number of items that should be returned
                the result set.
                This parameter is not supported for use in conjunction with the comment_id parameter.
                Acceptable values are 1 to 100, inclusive. The default value is 20.
            text_format:
                Whether the API should return comments formatted as HTML or as plain text.
                The default value is html.
                Acceptable values are:
                    - html: Returns the comments in HTML format.
                    - plainText: Returns the comments in plain text format.
            page_token:
                The parameter identifies a specific page in the result set that should be returned.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Comments data
        Raises:
            PyYouTubeException: Missing filter parameter.
        """

        params = {
            "part": enf_parts(resource="comments", value=parts),
            "maxResults": max_results,
            "textFormat": text_format,
            "pageToken": page_token,
            **kwargs,
        }
        if comment_id is not None:
            params["id"] = enf_comma_separated(field="comment_id", value=comment_id)
        elif parent_id is not None:
            params["parentId"] = parent_id
        else:
            raise PyYouTubeException(
                ErrorMessage(
                    status_code=ErrorCode.MISSING_PARAMS,
                    message=f"Specify at least one of for_username, id, managedByMe or mine",
                )
            )

        response = self._client.request(path="comments", params=params)
        data = self._client.parse_response(response=response)
        return data if return_json else CommentListResponse.from_dict(data)

    def insert(
        self,
        body: Union[dict, Comment],
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: bool = False,
        **kwargs,
    ) -> Union[dict, Comment]:
        """Creates a reply to an existing comment.

        Notes:
            To create a top-level comment, use the commentThreads.insert method.

        Args:
            body:
                Provide a comment resource in the request body. You can give dataclass or just a dict with data.
            parts:
                Comma-separated list of one or more comment resource properties.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Comment data.
        """

        params = {"part": enf_parts(resource="comments", value=parts), **kwargs}
        response = self._client.request(
            method="POST",
            path="comments",
            params=params,
            json=body,
        )
        data = self._client.parse_response(response=response)
        return data if return_json else Comment.from_dict(data)

    def update(
        self,
        body: Union[dict, Comment],
        parts: Optional[Union[str, list, tuple, set]] = None,
        return_json: bool = False,
        **kwargs,
    ) -> Union[dict, Comment]:
        """Modifies a comment.

        Args:
            body:
                Provide a comment resource in the request body. You can give dataclass or just a dict with data.
            parts:
                Comma-separated list of one or more comment resource properties.
            return_json:
                Type for returned data. If you set True JSON data will be returned.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Comment updated data.

        """
        params = {"part": enf_parts(resource="comments", value=parts), **kwargs}
        response = self._client.request(
            method="PUT",
            path="comments",
            params=params,
            json=body,
        )
        data = self._client.parse_response(response=response)
        return data if return_json else Comment.from_dict(data)

    def mark_as_spam(
        self,
        comment_id: str,
        **kwargs,
    ) -> bool:
        """Expresses the caller's opinion that one or more comments should be flagged as spam.

        Deprecated at [2023.09.12](https://developers.google.com/youtube/v3/revision_history#september-12,-2023)

        Args:
            comment_id:
                ID for the target comment.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Mark as spam status.

        """
        params = {"id": comment_id, **kwargs}
        response = self._client.request(
            method="POST",
            path="comments/markAsSpam",
            params=params,
        )
        if response.ok:
            return True
        self._client.parse_response(response=response)

    def set_moderation_status(
        self,
        comment_id: str,
        moderation_status: str,
        ban_author: Optional[bool] = None,
        **kwargs,
    ) -> bool:
        """Sets the moderation status of one or more comments.

        Args:
            comment_id:
                ID for the target comment.
            moderation_status:
                Identifies the new moderation status of the specified comments.
                Acceptable values:
                    - heldForReview: Marks a comment as awaiting review by a moderator.
                    - published: Clears a comment for public display.
                    - rejected:  Rejects a comment as being unfit for display.
                        This action also effectively hides all replies to the rejected comment.
            ban_author:
                Set the parameter value to true to ban the author.
                 This parameter is only valid if the moderationStatus parameter is also set to rejected.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Moderation status set status.
        """
        params = {
            "id": comment_id,
            "moderationStatus": moderation_status,
            "banAuthor": ban_author,
            **kwargs,
        }
        response = self._client.request(
            method="POST",
            path="comments/setModerationStatus",
            params=params,
        )
        if response.ok:
            return True
        self._client.parse_response(response=response)

    def delete(
        self,
        comment_id: str,
        **kwargs,
    ) -> bool:
        """Deletes a comment.

        Args:
            comment_id:
                ID for the target comment.
            **kwargs:
                Additional parameters for system parameters.
                Refer: https://cloud.google.com/apis/docs/system-parameters.

        Returns:
            Comment delete status.

        """
        params = {"id": comment_id, **kwargs}
        response = self._client.request(
            method="DELETE",
            path="comments",
            params=params,
        )
        if response.ok:
            return True
        self._client.parse_response(response=response)



"""
    These are member related models.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .common import BaseApiResponse
from .mixins import DatetimeTimeMixin


@dataclass
class MemberSnippetMemberDetails(BaseModel):
    """
    A class representing the member snippet member detail.

    Refer: https://developers.google.com/youtube/v3/docs/members#snippet.memberDetails
    """

    channelId: Optional[str] = field(default=None)
    channelUrl: Optional[str] = field(default=None, repr=False)
    displayName: Optional[str] = field(default=None, repr=False)
    profileImageUrl: Optional[str] = field(default=None, repr=False)


@dataclass
class MemberSnippetMembershipsDuration(BaseModel, DatetimeTimeMixin):
    memberSince: Optional[str] = field(default=None)
    memberTotalDurationMonths: Optional[int] = field(default=None, repr=False)


@dataclass
class MemberSnippetMembershipsDurationAtLevel(BaseModel):
    level: Optional[str] = field(default=None)
    memberSince: Optional[str] = field(default=None, repr=False)
    memberTotalDurationMonths: Optional[int] = field(default=None, repr=False)


@dataclass
class MemberSnippetMembershipsDetails(BaseModel):
    """
    A class representing the member snippet membership detail.

    Refer: https://developers.google.com/youtube/v3/docs/members#snippet.membershipsDetails
    """

    highestAccessibleLevel: Optional[str] = field(default=None)
    highestAccessibleLevelDisplayName: Optional[str] = field(default=None)
    accessibleLevels: Optional[List[str]] = field(default=None, repr=False)
    membershipsDuration: Optional[MemberSnippetMembershipsDuration] = field(
        default=None, repr=False
    )
    membershipsDurationAtLevel: Optional[
        List[MemberSnippetMembershipsDurationAtLevel]
    ] = field(default=None, repr=False)


@dataclass
class MemberSnippet(BaseModel):
    """
    A class representing the member snippet info.

    Refer: https://developers.google.com/youtube/v3/docs/members#snippet
    """

    creatorChannelId: Optional[str] = field(default=None)
    memberDetails: Optional[MemberSnippetMemberDetails] = field(
        default=None, repr=False
    )
    membershipsDetails: Optional[MemberSnippetMembershipsDetails] = field(
        default=None, repr=False
    )


@dataclass
class Member(BaseModel):
    """
    A class representing the member info.

    Refer: https://developers.google.com/youtube/v3/docs/members
    """

    kind: Optional[str] = field(default=None)
    etag: Optional[str] = field(default=None, repr=False)
    snippet: Optional[MemberSnippet] = field(default=None, repr=False)


@dataclass
class MemberListResponse(BaseApiResponse):
    """
    A class representing the member's retrieve response info.

    Refer: https://developers.google.com/youtube/v3/docs/members/list#response
    """

    items: Optional[List[Member]] = field(default=None, repr=False)



"""
    These are membership level related models.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .base import BaseModel
from .common import BaseResource, BaseApiResponse


@dataclass
class MembershipLevelSnippetLevelDetails(BaseModel):
    displayName: Optional[str] = field(default=None)


@dataclass
class MembershipsLevelSnippet(BaseModel):
    """
    A class representing the membership level snippet.

    Refer: https://developers.google.com/youtube/v3/docs/membershipsLevels#snippet
    """

    creatorChannelId: Optional[str] = field(default=None)
    levelDetails: Optional[MembershipLevelSnippetLevelDetails] = field(
        default=None, repr=False
    )


@dataclass
class MembershipsLevel(BaseResource):
    """
    A class representing the membership level.

    Refer: https://developers.google.com/youtube/v3/docs/membershipsLevels
    """

    snippet: Optional[MembershipsLevelSnippet] = field(default=None, repr=False)


@dataclass
class MembershipsLevelListResponse(BaseApiResponse):
    """
    A class representing the memberships level's retrieve response info.

    Refer: https://developers.google.com/youtube/v3/docs/membershipsLevels/list#response
    """

    items: Optional[List[MembershipsLevel]] = field(default=None, repr=False)
